import json
from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase

from marshmallow import ValidationError

from .......cache.base import BaseCache
from .......cache.in_memory import InMemoryCache
from .......indy.holder import IndyHolder
from .......indy.issuer import IndyIssuer
from .......ledger.base import BaseLedger
from .......ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .......messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from .......messaging.decorators.attach_decorator import AttachDecorator
from .......multitenant.base import BaseMultitenantManager
from .......multitenant.manager import MultitenantManager
from .......storage.base import BaseStorage
from .......storage.record import StorageRecord
from .......tests import mock
from .......utils.testing import create_test_profile
from ....message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ....messages.cred_format import V20CredFormat
from ....messages.cred_issue import V20CredIssue
from ....messages.cred_offer import V20CredOffer
from ....messages.cred_proposal import V20CredProposal
from ....messages.cred_request import V20CredRequest
from ....messages.inner.cred_preview import V20CredAttrSpec, V20CredPreview
from ....models.cred_ex_record import V20CredExRecord
from ....models.detail.indy import V20CredExRecordIndy
from ...handler import V20CredFormatError
from .. import handler as test_module
from ..handler import LOGGER as INDY_LOGGER
from ..handler import IndyCredFormatHandler

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "bc-reg"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
SCHEMA = {
    "ver": "1.0",
    "id": SCHEMA_ID,
    "name": SCHEMA_NAME,
    "version": "1.0",
    "attrNames": ["legalName", "jurisdictionId", "incorporationDate"],
    "seqNo": SCHEMA_TXN,
}
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:tag1"
CRED_DEF = {
    "ver": "1.0",
    "id": CRED_DEF_ID,
    "schemaId": SCHEMA_TXN,
    "type": "CL",
    "tag": "tag1",
    "value": {
        "primary": {
            "n": "...",
            "s": "...",
            "r": {
                "master_secret": "...",
                "legalName": "...",
                "jurisdictionId": "...",
                "incorporationDate": "...",
            },
            "rctxt": "...",
            "z": "...",
        },
        "revocation": {
            "g": "1 ...",
            "g_dash": "1 ...",
            "h": "1 ...",
            "h0": "1 ...",
            "h1": "1 ...",
            "h2": "1 ...",
            "htilde": "1 ...",
            "h_cap": "1 ...",
            "u": "1 ...",
            "pk": "1 ...",
            "y": "1 ...",
        },
    },
}
REV_REG_DEF_TYPE = "CL_ACCUM"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:{REV_REG_DEF_TYPE}:tag1"
TAILS_DIR = "/tmp/indy/revocation/tails_files"
TAILS_HASH = "8UW1Sz5cqoUnK9hqQk7nvtKK65t7Chu3ui866J23sFyJ"
TAILS_LOCAL = f"{TAILS_DIR}/{TAILS_HASH}"
REV_REG_DEF = {
    "ver": "1.0",
    "id": REV_REG_ID,
    "revocDefType": "CL_ACCUM",
    "tag": "tag1",
    "credDefId": CRED_DEF_ID,
    "value": {
        "issuanceType": "ISSUANCE_ON_DEMAND",
        "maxCredNum": 5,
        "publicKeys": {"accumKey": {"z": "1 ..."}},
        "tailsHash": TAILS_HASH,
        "tailsLocation": TAILS_LOCAL,
    },
}
INDY_OFFER = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "key_correctness_proof": {
        "c": "123467890",
        "xz_cap": "12345678901234567890",
        "xr_cap": [
            [
                "remainder",
                "1234567890",
            ],
            [
                "number",
                "12345678901234",
            ],
            [
                "master_secret",
                "12345678901234",
            ],
        ],
    },
    "nonce": "1234567890",
}
INDY_CRED_REQ = {
    "prover_did": TEST_DID,
    "cred_def_id": CRED_DEF_ID,
    "blinded_ms": {
        "u": "12345",
        "ur": "1 123467890ABCDEF",
        "hidden_attributes": ["master_secret"],
        "committed_attributes": {},
    },
    "blinded_ms_correctness_proof": {
        "c": "77777",
        "v_dash_cap": "12345678901234567890",
        "m_caps": {"master_secret": "271283714"},
        "r_caps": {},
    },
    "nonce": "9876543210",
}
INDY_CRED = {
    "schema_id": SCHEMA_ID,
    "cred_def_id": CRED_DEF_ID,
    "rev_reg_id": REV_REG_ID,
    "values": {
        "legalName": {
            "raw": "The Original House of Pies",
            "encoded": "108156129846915621348916581250742315326283968964",
        },
        "busId": {"raw": "11155555", "encoded": "11155555"},
        "jurisdictionId": {"raw": "1", "encoded": "1"},
        "incorporationDate": {
            "raw": "2021-01-01",
            "encoded": "121381685682968329568231",
        },
        "pic": {"raw": "cG90YXRv", "encoded": "125362825623562385689562"},
    },
    "signature": {
        "p_credential": {
            "m_2": "13683295623862356",
            "a": "1925723185621385238953",
            "e": "253516862326",
            "v": "26890295622385628356813632",
        },
        "r_credential": {
            "sigma": "1 00F81D",
            "c": "158698926BD09866E",
            "vr_prime_prime": "105682396DDF1A",
            "witness_signature": {"sigma_i": "1 ...", "u_i": "1 ...", "g_i": "1 ..."},
            "g_i": "1 ...",
            "i": 1,
            "m2": "862186285926592362384FA97FF3A4AB",
        },
    },
    "signature_correctness_proof": {
        "se": "10582965928638296868123",
        "c": "2816389562839651",
    },
    "rev_reg": {"accum": "21 ..."},
    "witness": {"omega": "21 ..."},
}


class TestV20IndyCredFormatHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

        # Ledger
        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=SCHEMA)
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=CRED_DEF)
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(return_value=REV_REG_DEF)
        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )
        # Context
        self.cache = InMemoryCache()
        self.profile.context.injector.bind_instance(BaseCache, self.cache)

        # Issuer
        self.issuer = mock.MagicMock(IndyIssuer, autospec=True)
        self.profile.context.injector.bind_instance(IndyIssuer, self.issuer)

        # Holder
        self.holder = mock.MagicMock(IndyHolder, autospec=True)
        self.profile.context.injector.bind_instance(IndyHolder, self.holder)

        self.handler = IndyCredFormatHandler(self.profile)
        assert self.handler.profile

    async def test_validate_fields(self):
        # Test correct data
        self.handler.validate_fields(CRED_20_PROPOSAL, {"cred_def_id": CRED_DEF_ID})
        self.handler.validate_fields(CRED_20_OFFER, INDY_OFFER)
        self.handler.validate_fields(CRED_20_REQUEST, INDY_CRED_REQ)
        self.handler.validate_fields(CRED_20_ISSUE, INDY_CRED)

        # test incorrect proposal
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(
                CRED_20_PROPOSAL, {"some_random_key": "some_random_value"}
            )

        # test incorrect offer
        with self.assertRaises(ValidationError):
            offer = INDY_OFFER.copy()
            offer.pop("nonce")
            self.handler.validate_fields(CRED_20_OFFER, offer)

        # test incorrect request
        with self.assertRaises(ValidationError):
            req = INDY_CRED_REQ.copy()
            req.pop("nonce")
            self.handler.validate_fields(CRED_20_REQUEST, req)

        # test incorrect cred
        with self.assertRaises(ValidationError):
            cred = INDY_CRED.copy()
            cred.pop("schema_id")
            self.handler.validate_fields(CRED_20_ISSUE, cred)

    async def test_get_indy_detail_record(self):
        cred_ex_id = "dummy"
        details_indy = [
            V20CredExRecordIndy(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="0",
            ),
            V20CredExRecordIndy(
                cred_ex_id=cred_ex_id,
                rev_reg_id="rr-id",
                cred_rev_id="1",
            ),
        ]
        async with self.profile.session() as session:
            await details_indy[0].save(session)
            await details_indy[1].save(session)  # exercise logger warning on get()

        with mock.patch.object(INDY_LOGGER, "warning", mock.MagicMock()) as mock_warning:
            assert await self.handler.get_detail_record(cred_ex_id) in details_indy
            mock_warning.assert_called_once()

    async def test_check_uniqueness(self):
        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_indy_query:
            mock_indy_query.return_value = []
            await self.handler._check_uniqueness("dummy-cx-id")

        with mock.patch.object(
            self.handler.format.detail,
            "query_by_cred_ex_id",
            mock.CoroutineMock(),
        ) as mock_indy_query:
            mock_indy_query.return_value = [mock.MagicMock()]
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler._check_uniqueness("dummy-cx-id")
            assert "detail record already exists" in str(context.exception)

    async def test_create_proposal(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = {"schema_id": SCHEMA_ID}

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == proposal_data

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_proposal_none(self):
        cred_ex_record = mock.MagicMock()
        proposal_data = None

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, proposal_data
        )

        # assert content of attachment is proposal data
        assert attachment.content == {}

    async def test_receive_proposal(self):
        cred_ex_record = mock.MagicMock()
        cred_proposal_message = mock.MagicMock()

        # Not much to assert. Receive proposal doesn't do anything
        await self.handler.receive_proposal(cred_ex_record, cred_proposal_message)

    async def test_create_offer(self):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(
                V20CredAttrSpec(name="legalName", value="value"),
                V20CredAttrSpec(name="jurisdictionId", value="value"),
                V20CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(cred_def_record)

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(INDY_OFFER)
        )

        (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

        self.issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == INDY_OFFER

        # assert data is encoded as base64
        assert attachment.data.base64

        self.issuer.create_credential_offer.reset_mock()
        await self.handler.create_offer(cred_proposal)
        self.issuer.create_credential_offer.assert_not_called()

    async def test_create_offer_no_cache(self):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(
                V20CredAttrSpec(name="legalName", value="value"),
                V20CredAttrSpec(name="jurisdictionId", value="value"),
                V20CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )

        # Remove cache from injection context
        self.profile.context.injector.clear_binding(BaseCache)

        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(cred_def_record)

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(INDY_OFFER)
        )

        (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

        self.issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == INDY_OFFER

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_offer_attr_mismatch(self):
        schema_id_parts = SCHEMA_ID.split(":")

        cred_preview = V20CredPreview(
            attributes=(  # names have spaces instead of camel case
                V20CredAttrSpec(name="legal name", value="value"),
                V20CredAttrSpec(name="jurisdiction id", value="value"),
                V20CredAttrSpec(name="incorporation date", value="value"),
            )
        )

        cred_proposal = V20CredProposal(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64({"cred_def_id": CRED_DEF_ID}, ident="0")
            ],
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )

        cred_def_record = StorageRecord(
            CRED_DEF_SENT_RECORD_TYPE,
            CRED_DEF_ID,
            {
                "schema_id": SCHEMA_ID,
                "schema_issuer_did": schema_id_parts[0],
                "schema_name": schema_id_parts[-2],
                "schema_version": schema_id_parts[-1],
                "issuer_did": TEST_DID,
                "cred_def_id": CRED_DEF_ID,
                "epoch": str(int(time())),
            },
        )
        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            await storage.add_record(cred_def_record)

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(INDY_OFFER)
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=(None, self.ledger)),
        ):
            with self.assertRaises(V20CredFormatError):
                await self.handler.create_offer(cred_proposal)

    async def test_create_offer_no_matching_sent_cred_def(self):
        cred_proposal = V20CredProposal(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64({}, ident="0")],
        )

        self.issuer.create_credential_offer = mock.CoroutineMock(
            return_value=json.dumps(INDY_OFFER)
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_offer(cred_proposal)
        assert "Issuer has no operable cred def" in str(context.exception)

    async def test_receive_offer(self):
        cred_ex_record = mock.MagicMock()
        cred_offer_message = mock.MagicMock()

        # Not much to assert. Receive offer doesn't do anything
        await self.handler.receive_offer(cred_ex_record, cred_offer_message)

    async def test_create_request(self):
        holder_did = "did"

        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer.serialize(),
        )

        cred_def = {"cred": "def"}
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=cred_def)

        cred_req_meta = {}
        self.holder.create_credential_request = mock.CoroutineMock(
            return_value=(json.dumps(INDY_CRED_REQ), json.dumps(cred_req_meta))
        )

        (cred_format, attachment) = await self.handler.create_request(
            cred_ex_record, {"holder_did": holder_did}
        )

        self.holder.create_credential_request.assert_called_once_with(
            INDY_OFFER, cred_def, holder_did
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == INDY_CRED_REQ

        # assert data is encoded as base64
        assert attachment.data.base64

        # cover case with cache (change ID to prevent already exists error)
        cred_ex_record._id = "dummy-id2"
        await self.handler.create_request(cred_ex_record, {"holder_did": holder_did})

        # cover case with no cache in injection context
        self.profile.context.injector.clear_binding(BaseCache)
        cred_ex_record._id = "dummy-id3"
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=(None, self.ledger)),
        ):
            await self.handler.create_request(cred_ex_record, {"holder_did": holder_did})

    async def test_create_request_bad_state(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_OFFER_SENT)

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_request(cred_ex_record)
        assert "Indy issue credential format cannot start from credential request" in str(
            context.exception
        )

        cred_ex_record.state = None

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_request(cred_ex_record)
        assert "Indy issue credential format cannot start from credential request" in str(
            context.exception
        )

    async def test_create_request_not_unique_x(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_OFFER_RECEIVED)

        with mock.patch.object(
            self.handler, "_check_uniqueness", mock.CoroutineMock()
        ) as mock_unique:
            mock_unique.side_effect = (
                V20CredFormatError("indy detail record already exists"),
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.create_request(cred_ex_record)

            assert "indy detail record already exists" in str(context.exception)

    async def test_receive_request(self):
        cred_ex_record = mock.MagicMock()
        cred_request_message = mock.MagicMock()

        # Not much to assert. Receive request doesn't do anything
        await self.handler.receive_request(cred_ex_record, cred_request_message)

    async def test_receive_request_no_offer(self):
        cred_ex_record = mock.MagicMock(cred_offer=None)
        cred_request_message = mock.MagicMock()

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_request(cred_ex_record, cred_request_message)

        assert "Indy issue credential format cannot start from credential request" in str(
            context.exception
        )

    async def test_issue_credential_revocable(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        cred_rev_id = "1000"
        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(INDY_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=(
                    mock.MagicMock(  # active_rev_reg_rec
                        revoc_reg_id=REV_REG_ID,
                    ),
                    mock.MagicMock(  # rev_reg
                        tails_local_path="dummy-path",
                        get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                        max_creds=10,
                    ),
                )
            )

            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=1
            )

            self.issuer.create_credential.assert_called_once_with(
                SCHEMA,
                INDY_OFFER,
                INDY_CRED_REQ,
                attr_values,
                REV_REG_ID,
                "dummy-path",
            )

            # assert identifier match
            assert cred_format.attach_id == self.handler.format.api == attachment.ident

            # assert content of attachment is proposal data
            assert attachment.content == INDY_CRED

            # assert data is encoded as base64
            assert attachment.data.base64

    async def test_issue_credential_non_revocable(self):
        CRED_DEF_NR = deepcopy(CRED_DEF)
        CRED_DEF_NR["value"]["revocation"] = None
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(INDY_CRED), None)
        )
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value=CRED_DEF_NR
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with mock.patch.object(
            IndyLedgerRequestsExecutor,
            "get_ledger_for_identifier",
            mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
        ):
            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record, retries=0
            )

            self.issuer.create_credential.assert_called_once_with(
                SCHEMA,
                INDY_OFFER,
                INDY_CRED_REQ,
                attr_values,
                None,
                None,
            )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == INDY_CRED

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_issue_credential_not_unique_x(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_REQUEST_RECEIVED)

        with mock.patch.object(
            self.handler, "_check_uniqueness", mock.CoroutineMock()
        ) as mock_unique:
            mock_unique.side_effect = (
                V20CredFormatError("indy detail record already exists"),
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record)

            assert "indy detail record already exists" in str(context.exception)

    async def test_issue_credential_no_active_rr_no_retries(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_rev_id = "1"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(INDY_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=()
            )
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=0)
            assert "has no active revocation registry" in str(context.exception)

    async def test_issue_credential_no_active_rr_retry(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_rev_id = "1"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(INDY_CRED), cred_rev_id)
        )

        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                side_effect=[
                    None,
                    (
                        mock.MagicMock(  # active_rev_reg_rec
                            revoc_reg_id=REV_REG_ID,
                            set_state=mock.CoroutineMock(),
                        ),
                        mock.MagicMock(  # rev_reg
                            tails_local_path="dummy-path",
                            get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                        ),
                    ),
                ]
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=1)
            assert "has no active revocation registry" in str(context.exception)

    async def test_issue_credential_rr_full(self):
        attr_values = {
            "legalName": "value",
            "jurisdictionId": "value",
            "incorporationDate": "value",
        }
        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v) for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            initiator=V20CredExRecord.INITIATOR_SELF,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_REQUEST_RECEIVED,
        )

        self.issuer.create_credential = mock.CoroutineMock(
            side_effect=test_module.IndyIssuerRevocationRegistryFullError("Nope")
        )
        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=(
                    mock.MagicMock(  # active_rev_reg_rec
                        revoc_reg_id=REV_REG_ID,
                        set_state=mock.CoroutineMock(),
                    ),
                    mock.MagicMock(  # rev_reg
                        tails_local_path="dummy-path",
                        get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                    ),
                )
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.issue_credential(cred_ex_record, retries=1)
            assert "has no active revocation registry" in str(context.exception)

    async def test_receive_credential(self):
        cred_ex_record = mock.MagicMock()
        cred_issue_message = mock.MagicMock()

        # Not much to assert. Receive credential doesn't do anything
        await self.handler.receive_credential(cred_ex_record, cred_issue_message)

    async def test_store_credential(self):
        connection_id = "test_conn_id"
        attr_values = {
            "legalName": ["value", None],
            "jurisdictionId": ["value", None],
            "incorporationDate": ["value", None],
            "pic": ["cG90YXRv", "image/jpeg"],
        }
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v[0], mime_type=v[1])
                for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_offer.assign_thread_id(thread_id)
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(INDY_CRED, ident="0")],
        )

        stored_cx_rec = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            connection_id=connection_id,
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            cred_issue=cred_issue.serialize(),
            initiator=V20CredExRecord.INITIATOR_EXTERNAL,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            auto_remove=True,
        )

        cred_id = "cred-id"

        self.holder.store_credential = mock.CoroutineMock(return_value=cred_id)
        stored_cred = {"stored": "cred"}
        self.holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(stored_cred)
        )

        with mock.patch.object(
            test_module, "RevocationRegistry", autospec=True
        ) as mock_rev_reg:
            mock_rev_reg.from_definition = mock.MagicMock(
                return_value=mock.MagicMock(
                    get_or_fetch_local_tails_path=mock.CoroutineMock()
                )
            )
            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.store_credential(stored_cx_rec, cred_id=cred_id)
            assert "No credential exchange " in str(context.exception)
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        with (
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
            mock.patch.object(
                test_module, "RevocationRegistry", autospec=True
            ) as mock_rev_reg,
            mock.patch.object(
                test_module.IndyCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_get_detail_record,
        ):
            mock_rev_reg.from_definition = mock.MagicMock(
                return_value=mock.MagicMock(
                    get_or_fetch_local_tails_path=mock.CoroutineMock()
                )
            )
            mock_get_detail_record.return_value = mock.MagicMock(
                cred_request_metadata=cred_req_meta,
                save=mock.CoroutineMock(),
            )

            self.ledger.get_credential_definition.reset_mock()
            await self.handler.store_credential(stored_cx_rec, cred_id=cred_id)

            self.ledger.get_credential_definition.assert_called_once_with(CRED_DEF_ID)

            self.holder.store_credential.assert_called_once_with(
                CRED_DEF,
                INDY_CRED,
                cred_req_meta,
                {"pic": "image/jpeg"},
                credential_id=cred_id,
                rev_reg_def=REV_REG_DEF,
            )

    async def test_store_credential_holder_store_indy_error(self):
        connection_id = "test_conn_id"
        attr_values = {
            "legalName": ["value", None],
            "jurisdictionId": ["value", None],
            "incorporationDate": ["value", None],
            "pic": ["cG90YXRv", "image/jpeg"],
        }
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        cred_preview = V20CredPreview(
            attributes=[
                V20CredAttrSpec(name=k, value=v[0], mime_type=v[1])
                for (k, v) in attr_values.items()
            ]
        )
        cred_offer = V20CredOffer(
            credential_preview=cred_preview,
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(INDY_OFFER, ident="0")],
        )
        cred_offer.assign_thread_id(thread_id)
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(INDY_CRED_REQ, ident="0")],
        )
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.INDY.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(INDY_CRED, ident="0")],
        )

        stored_cx_rec = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            connection_id=connection_id,
            cred_offer=cred_offer.serialize(),
            cred_request=cred_request.serialize(),
            cred_issue=cred_issue.serialize(),
            initiator=V20CredExRecord.INITIATOR_EXTERNAL,
            role=V20CredExRecord.ROLE_ISSUER,
            state=V20CredExRecord.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            auto_remove=True,
        )

        cred_id = "cred-id"
        self.holder.store_credential = mock.CoroutineMock(
            side_effect=test_module.IndyHolderError("Problem", {"message": "Nope"})
        )

        with (
            mock.patch.object(
                test_module.IndyCredFormatHandler, "get_detail_record", autospec=True
            ) as mock_get_detail_record,
            mock.patch.object(
                test_module.RevocationRegistry, "from_definition", mock.MagicMock()
            ) as mock_rev_reg,
        ):
            mock_get_detail_record.return_value = mock.MagicMock(
                cred_request_metadata=cred_req_meta,
                save=mock.CoroutineMock(),
            )
            mock_rev_reg.return_value = mock.MagicMock(
                get_or_fetch_local_tails_path=mock.CoroutineMock()
            )
            with self.assertRaises(test_module.IndyHolderError) as context:
                await self.handler.store_credential(stored_cx_rec, cred_id)
            assert "Nope" in str(context.exception)



import json
from base64 import b64encode
from unittest import IsolatedAsyncioTestCase

from didcomm_messaging import DIDCommMessaging, PackResult
from didcomm_messaging.crypto.backend.askar import CryptoServiceError

from ...protocols.didcomm_prefix import DIDCommPrefix
from ...protocols.routing.v1_0.message_types import FORWARD
from ...tests import mock
from ...transport.inbound.receipt import MessageReceipt
from ...transport.v2_pack_format import V2PackWireFormat
from ...utils.testing import create_test_profile
from ...wallet.base import BaseWallet
from ...wallet.did_method import SOV, DIDMethods
from ...wallet.error import WalletError
from ...wallet.key_type import ED25519
from .. import pack_format as test_module
from ..error import RecipientKeysError, WireFormatEncodeError, WireFormatParseError
from ..pack_format import PackWireFormat


class TestPackWireFormat(IsolatedAsyncioTestCase):
    test_message_type = DIDCommPrefix.qualify_current("PROTOCOL/MESSAGE")
    test_message_id = "MESSAGE_ID"
    test_content = "CONTENT"
    test_thread_id = "THREAD_ID"
    test_message = {
        "@type": test_message_type,
        "@id": test_message_id,
        "~thread": {"thid": test_thread_id},
        "~transport": {"return_route": "all"},
        "content": test_content,
    }
    test_seed = "testseed000000000000000000000001"
    test_routing_seed = "testseed000000000000000000000002"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())

    async def test_errors(self):
        serializer = PackWireFormat()
        bad_values = [None, "", "1", "[]", "{..."]

        async with self.profile.session() as session:
            for message_json in bad_values:
                with self.assertRaises(WireFormatParseError):
                    await serializer.parse_message(session, message_json)

            x_message = {
                "@id": TestPackWireFormat.test_message_id,
                "~thread": {"thid": TestPackWireFormat.test_thread_id},
                "~transport": {"return_route": "all"},
                "content": "{}",
            }

            serializer.task_queue = None
            with mock.patch.object(
                serializer.v1pack_format, "unpack", mock.CoroutineMock()
            ) as mock_unpack:
                mock_unpack.return_value = "{missing-brace"
                with self.assertRaises(WireFormatParseError) as context:
                    await serializer.parse_message(session, json.dumps(x_message))
            assert "Message JSON parsing failed" in str(context.exception)

            serializer = PackWireFormat()
            serializer.task_queue = None
            with mock.patch.object(
                serializer.v1pack_format, "unpack", mock.CoroutineMock()
            ) as mock_unpack:
                mock_unpack.return_value = json.dumps([1, 2, 3])
                with self.assertRaises(WireFormatParseError) as context:
                    await serializer.parse_message(session, json.dumps(x_message))
            assert "Message JSON result is not an object" in str(context.exception)

            with self.assertRaises(WireFormatParseError):
                await serializer.unpack(session, "...", MessageReceipt())

    async def test_pack_x(self):
        serializer = PackWireFormat()

        async with self.profile.session() as session:
            with self.assertRaises(WireFormatEncodeError):
                await serializer.pack(session, None, [], [], "key")
            with self.assertRaises(WireFormatEncodeError):
                await serializer.pack(
                    session,
                    None,
                    ["key"],
                    None,
                    ["key"],
                )

        mock_wallet = mock.MagicMock(
            pack_message=mock.CoroutineMock(side_effect=WalletError())
        )
        self.profile.context.injector.bind_instance(BaseWallet, mock_wallet)
        async with self.profile.session() as session:
            with self.assertRaises(WireFormatEncodeError):
                await serializer.pack(session, None, ["key"], None, ["key"])

        mock_wallet = mock.MagicMock(
            pack_message=mock.CoroutineMock(
                side_effect=[json.dumps("message").encode("utf-8"), WalletError()]
            )
        )
        self.profile.context.injector.bind_instance(BaseWallet, mock_wallet)
        async with self.profile.session() as session:
            with mock.patch.object(
                test_module, "Forward", mock.MagicMock()
            ) as mock_forward:
                mock_forward.return_value = mock.MagicMock(to_json=mock.MagicMock())
                with self.assertRaises(WireFormatEncodeError):
                    await serializer.pack(session, None, ["key"], ["key"], ["key"])

    async def test_unpacked(self):
        serializer = PackWireFormat()
        message_json = json.dumps(self.test_message)
        async with self.profile.session() as session:
            message_dict, delivery = await serializer.parse_message(session, message_json)
            assert message_dict == self.test_message
            assert message_dict["@type"] == self.test_message_type
            assert delivery.thread_id == self.test_thread_id
            assert delivery.direct_response_mode == "all"

    async def test_fallback(self):
        serializer = PackWireFormat()

        message = self.test_message.copy()
        message.pop("@type")
        message_json = json.dumps(message)

        async with self.profile.session() as session:
            message_dict, delivery = await serializer.parse_message(session, message_json)
            assert delivery.raw_message == message_json
            assert message_dict == message

    async def test_encode_decode(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            local_did = await wallet.create_local_did(
                method=SOV, key_type=ED25519, seed=self.test_seed
            )
            serializer = PackWireFormat()
            recipient_keys = (local_did.verkey,)
            routing_keys = ()
            sender_key = local_did.verkey
            message_json = json.dumps(self.test_message)

            packed_json = await serializer.encode_message(
                session, message_json, recipient_keys, routing_keys, sender_key
            )
            packed = json.loads(packed_json)

            assert isinstance(packed, dict) and "protected" in packed
            assert serializer.get_recipient_keys(packed_json) == list(recipient_keys)
            with self.assertRaises(test_module.RecipientKeysError):
                serializer.get_recipient_keys(message_json)

            message_dict, delivery = await serializer.parse_message(session, packed_json)
            assert message_dict == self.test_message
            assert message_dict["@type"] == self.test_message_type
            assert delivery.thread_id == self.test_thread_id
            assert delivery.direct_response_mode == "all"

            plain_json = json.dumps("plain")
            assert (
                await serializer.encode_message(session, plain_json, None, None, None)
                == plain_json
            )

    async def test_forward(self):
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            local_did = await wallet.create_local_did(
                method=SOV, key_type=ED25519, seed=self.test_seed
            )
            router_did = await wallet.create_local_did(
                method=SOV, key_type=ED25519, seed=self.test_routing_seed
            )
            serializer = PackWireFormat()
            recipient_keys = (local_did.verkey,)
            routing_keys = (router_did.verkey,)
            sender_key = local_did.verkey
            message_json = json.dumps(self.test_message)

            packed_json = await serializer.encode_message(
                session, message_json, recipient_keys, routing_keys, sender_key
            )
            packed = json.loads(packed_json)

            assert isinstance(packed, dict) and "protected" in packed

            message_dict, delivery = await serializer.parse_message(session, packed_json)
            assert message_dict["@type"] == DIDCommPrefix.qualify_current(FORWARD)
            assert delivery.recipient_verkey == router_did.verkey
            assert delivery.sender_verkey is None

    async def test_get_recipient_keys(self):
        recip_keys = ["kid1", "kid2", "kid3"]
        enc_message = {
            "protected": b64encode(
                json.dumps(
                    {"recipients": [{"header": {"kid": k}} for k in recip_keys]}
                ).encode("utf-8")
            ).decode()
        }

        serializer = PackWireFormat()
        actual_recip_keys = serializer.get_recipient_keys(json.dumps(enc_message))

        self.assertEqual(recip_keys, actual_recip_keys)

    async def test_get_recipient_keys_fails(self):
        enc_message = {"protected": {}}

        serializer = PackWireFormat()

        with self.assertRaises(RecipientKeysError):
            serializer.get_recipient_keys(json.dumps(enc_message))


class MockDIDCommMessaging(DIDCommMessaging):
    def __init__(
        self,
    ):
        crypto = mock.MagicMock()
        secrets = mock.MagicMock()
        resolver = mock.MagicMock()
        packaging = mock.AsyncMock()
        routing = mock.MagicMock()

        super().__init__(crypto, secrets, resolver, packaging, routing)


class TestV2PackWireFormat(IsolatedAsyncioTestCase):
    test_message = {
        "type": "test_message_type",
        "id": "test_message_id",
        "thid": "test_thread_id",
        "content": "test_content",
    }

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(DIDMethods, DIDMethods())
        self.profile.context.settings["experiment.didcomm_v2"] = True

    async def test_errors(self):
        wire_format = PackWireFormat()

        message = self.test_message.copy()
        message.pop("type")
        message_json = json.dumps(message)
        async with self.profile.session() as session:
            session.context.injector.bind_instance(
                DIDCommMessaging, MockDIDCommMessaging()
            )
            with self.assertRaises(WireFormatParseError) as context:
                await wire_format.parse_message(session, message_json)
            assert "Unable to determine appropriate WireFormat version" in str(
                context.exception
            )

            wire_format = V2PackWireFormat()
            bad_values = [None, "", "1", "[]", "{..."]

            for message_json in bad_values:
                with self.assertRaises(WireFormatParseError):
                    await wire_format.parse_message(session, message_json)

    async def test_fallback(self):
        serializer = V2PackWireFormat()

        test_dm = MockDIDCommMessaging()
        test_dm.packaging.unpack = mock.AsyncMock(side_effect=CryptoServiceError())
        async with self.profile.session() as session:
            session.context.injector.bind_instance(DIDCommMessaging, test_dm)

            message = self.test_message.copy()
            message.pop("type")
            message_json = json.dumps(message)

            message_dict, delivery = await serializer.parse_message(session, message_json)
            assert delivery.raw_message == message_json
            assert message_dict == message

    async def test_encode(self):
        serializer = V2PackWireFormat()

        test_dm = MockDIDCommMessaging()
        test_dm.pack = mock.AsyncMock(
            return_value=PackResult(
                message=self.test_message, target_services=mock.MagicMock()
            )
        )
        async with self.profile.session() as session:
            session.context.injector.bind_instance(DIDCommMessaging, test_dm)
            message = await serializer.encode_message(
                session=session,
                message_json=self.test_message,
                sender_key="sender_key",
                recipient_keys=["recip_key"],
                routing_keys=["route_key"],
            )

        assert message == self.test_message



"""Represents a forward message."""

import json
from typing import Optional, Union

from marshmallow import EXCLUDE, fields, pre_load

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from ..message_types import FORWARD, PROTOCOL_PACKAGE

HANDLER_CLASS = f"{PROTOCOL_PACKAGE}.handlers.forward_handler.ForwardHandler"


class Forward(AgentMessage):
    """Represents a request to forward a message to a connected agent."""

    class Meta:
        """Forward metadata."""

        handler_class = HANDLER_CLASS
        message_type = FORWARD
        schema_class = "ForwardSchema"

    def __init__(
        self, *, to: Optional[str] = None, msg: Union[dict, str] = None, **kwargs
    ):
        """Initialize forward message object.

        Args:
            to (str): Recipient DID
            msg (str): Message content
            kwargs: Additional keyword arguments
        """
        super().__init__(**kwargs)
        self.to = to
        if isinstance(msg, str):
            msg = json.loads(msg)
        self.msg = msg


class ForwardSchema(AgentMessageSchema):
    """Forward message schema used in serialization/deserialization."""

    class Meta:
        """ForwardSchema metadata."""

        model_class = Forward
        unknown = EXCLUDE

    @pre_load
    def handle_str_message(self, data, **kwargs):
        """Accept string value for msg, as produced by previous implementation."""
        if "msg" in data and isinstance(data["msg"], str):
            data["msg"] = json.loads(data["msg"])
        return data

    to = fields.Str(required=True)
    msg = fields.Dict(required=True)



from datetime import datetime, timezone
from unittest import IsolatedAsyncioTestCase

import pytest

from ....did.did_key import DIDKey
from ....resolver.default.key import KeyDIDResolver
from ....resolver.did_resolver import DIDResolver
from ....utils.testing import create_test_profile
from ....wallet.base import BaseWallet
from ....wallet.key_type import BLS12381G2, ED25519
from ...ld_proofs import (
    AssertionProofPurpose,
    BbsBlsSignature2020,
    BbsBlsSignatureProof2020,
    Ed25519Signature2018,
    Ed25519Signature2020,
    WalletKeyPair,
    derive,
    sign,
    verify,
)
from ...tests.document_loader import custom_document_loader
from ..document_loader import DocumentLoader
from .test_doc import (
    DOC_DERIVED_BBS,
    DOC_FRAME_BBS,
    DOC_SIGNED,
    DOC_SIGNED_2020,
    DOC_SIGNED_BBS,
    DOC_TEMPLATE,
    DOC_TEMPLATE_2020,
    DOC_TEMPLATE_BBS,
    DOC_VERIFIED,
    DOC_VERIFIED_2020,
)


class TestLDProofs(IsolatedAsyncioTestCase):
    test_seed = "testseed000000000000000000000001"

    async def asyncSetUp(self):
        self.profile = await create_test_profile()
        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)

            self.ed25519_key_info = await wallet.create_signing_key(
                key_type=ED25519, seed=self.test_seed
            )
            self.ed25519_verification_method = DIDKey.from_public_key_b58(
                self.ed25519_key_info.verkey, ED25519
            ).key_id

            self.bls12381g2_key_info = await wallet.create_signing_key(
                key_type=BLS12381G2, seed=self.test_seed
            )

            self.bls12381g2_verification_method = DIDKey.from_public_key_b58(
                self.bls12381g2_key_info.verkey, BLS12381G2
            ).key_id

        self.profile.context.injector.bind_instance(
            DIDResolver, DIDResolver([KeyDIDResolver()])
        )

    async def test_sign_Ed25519Signature2018(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = Ed25519Signature2018(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime(2019, 12, 11, 3, 50, 55, 0, timezone.utc),
        )
        signed = await sign(
            document=DOC_TEMPLATE,
            suite=suite,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert signed == DOC_SIGNED

    async def test_sign_Ed25519Signature2020(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = Ed25519Signature2020(
            verification_method=self.ed25519_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=ED25519,
                public_key_base58=self.ed25519_key_info.verkey,
            ),
            date=datetime(2019, 12, 11, 3, 50, 55, 0, timezone.utc),
        )
        signed = await sign(
            document=DOC_TEMPLATE_2020,
            suite=suite,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert signed == DOC_SIGNED_2020

    async def test_verify_Ed25519Signature2018(self):
        # Verification requires lot less input parameters
        suite = Ed25519Signature2018(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )

        result = await verify(
            document=DOC_SIGNED,
            suites=[suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result == DOC_VERIFIED

    async def test_verify_Ed25519Signature2020(self):
        # Verification requires lot less input parameters
        suite = Ed25519Signature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=ED25519),
        )

        result = await verify(
            document=DOC_SIGNED_2020,
            suites=[suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result == DOC_VERIFIED_2020

    @pytest.mark.ursa_bbs_signatures
    async def test_sign_BbsBlsSignature2020(self):
        # Use different key pair and suite for signing and verification
        # as during verification a lot of information can be extracted
        # from the proof / document
        suite = BbsBlsSignature2020(
            verification_method=self.bls12381g2_verification_method,
            key_pair=WalletKeyPair(
                profile=self.profile,
                key_type=BLS12381G2,
                public_key_base58=self.bls12381g2_key_info.verkey,
            ),
            date=datetime(2019, 12, 11, 3, 50, 55, 0),
        )
        signed = await sign(
            document=DOC_TEMPLATE_BBS,
            suite=suite,
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        # BBS generates unique signature every time so we cant compare signatures
        assert signed

        result = await verify(
            document=signed,
            suites=[suite],
            purpose=AssertionProofPurpose(),
            document_loader=DocumentLoader(self.profile),
        )

        assert result.verified

    @pytest.mark.ursa_bbs_signatures
    async def test_verify_BbsBlsSignature2020(self):
        # Verification requires lot less input parameters
        suite = BbsBlsSignature2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2),
        )

        result = await verify(
            document=DOC_SIGNED_BBS,
            suites=[suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified

    @pytest.mark.ursa_bbs_signatures
    async def test_derive_BbsBlsSignatureProof2020(self):
        # Verification requires lot less input parameters
        suite = BbsBlsSignatureProof2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2),
        )

        result = await derive(
            document=DOC_SIGNED_BBS,
            reveal_document=DOC_FRAME_BBS,
            suite=suite,
            document_loader=custom_document_loader,
        )

        assert result

    @pytest.mark.ursa_bbs_signatures
    async def test_verify_BbsBlsSignatureProof2020(self):
        # Verification requires lot less input parameters
        suite = BbsBlsSignatureProof2020(
            key_pair=WalletKeyPair(profile=self.profile, key_type=BLS12381G2),
        )

        result = await verify(
            document=DOC_DERIVED_BBS,
            suites=[suite],
            purpose=AssertionProofPurpose(),
            document_loader=custom_document_loader,
        )

        assert result.verified



"""JSON-LD, Linked Data Proof and Verifiable Credential constants."""

SECURITY_CONTEXT_V1_URL = "https://w3id.org/security/v1"
SECURITY_CONTEXT_V2_URL = "https://w3id.org/security/v2"
SECURITY_CONTEXT_V3_URL = "https://w3id.org/security/v3-unstable"
SECURITY_CONTEXT_URL = SECURITY_CONTEXT_V2_URL
DID_V1_CONTEXT_URL = "https://www.w3.org/ns/did/v1"
CREDENTIALS_CONTEXT_V1_URL = "https://www.w3.org/2018/credentials/v1"
CREDENTIALS_CONTEXT_V2_URL = "https://www.w3.org/ns/credentials/v2"
SECURITY_CONTEXT_BBS_URL = "https://w3id.org/security/bbs/v1"
SECURITY_CONTEXT_ED25519_2020_URL = "https://w3id.org/security/suites/ed25519-2020/v1"
SECURITY_CONTEXT_MULTIKEY_URL = "https://w3id.org/security/multikey/v1"

CREDENTIALS_ISSUER_URL = "https://www.w3.org/2018/credentials#issuer"
SECURITY_PROOF_URL = "https://w3id.org/security#proof"
SECURITY_SIGNATURE_URL = "https://w3id.org/security#signature"

VERIFIABLE_CREDENTIAL_TYPE = "VerifiableCredential"
VERIFIABLE_PRESENTATION_TYPE = "VerifiablePresentation"

EXPANDED_TYPE_CREDENTIALS_CONTEXT_V1_VC_TYPE = (
    "https://www.w3.org/2018/credentials#VerifiableCredential"
)



"""Test PeerDIDResolver."""

import pytest
import pytest_asyncio
from did_peer_2 import peer2to3

from ....connections.models.conn_record import ConnRecord
from ....core.event_bus import EventBus
from ....core.profile import Profile
from ....utils.testing import create_test_profile
from .. import peer3 as test_module
from ..peer2 import PeerDID2Resolver
from ..peer3 import PeerDID3Resolver

TEST_DP2 = "did:peer:2.Ez6LSpkcni2KTTxf4nAp6cPxjRbu26Tj4b957BgHcknVeNFEj.Vz6MksXhfmxm2i3RnoHH2mKQcx7EY4tToJR9JziUs6bp8a6FM.SeyJ0IjoiZGlkLWNvbW11bmljYXRpb24iLCJzIjoiaHR0cDovL2hvc3QuZG9ja2VyLmludGVybmFsOjkwNzAiLCJyZWNpcGllbnRfa2V5cyI6W119"

TEST_DP3 = peer2to3(TEST_DP2)


@pytest.fixture
def event_bus():
    yield EventBus()


@pytest_asyncio.fixture
async def profile(event_bus: EventBus):
    """Profile fixture."""
    profile = await create_test_profile()
    profile.context.injector.bind_instance(EventBus, event_bus)
    yield profile


@pytest_asyncio.fixture
async def resolver(profile):
    """Resolver fixture."""
    instance = PeerDID3Resolver()
    await instance.setup(profile.context)
    yield instance


@pytest.fixture
def peer2_resolver():
    """Resolver fixture."""
    yield PeerDID2Resolver()


@pytest.mark.asyncio
async def test_resolve_2_then_3(
    profile: Profile, resolver: PeerDID3Resolver, peer2_resolver: PeerDID2Resolver
):
    """Test resolver setup."""
    assert resolver.supported_did_regex
    assert await resolver.supports(profile, TEST_DP3)

    await peer2_resolver.resolve(profile, TEST_DP2)
    doc = await resolver.resolve(profile, TEST_DP3)

    assert doc
    assert doc["id"] == TEST_DP3
    assert "service" in doc
    assert "verificationMethod" in doc
    assert len(doc["verificationMethod"]) == 2


@pytest.mark.asyncio
async def test_resolve_x_no_2(profile: Profile, resolver: PeerDID3Resolver):
    """Test resolver setup."""
    with pytest.raises(test_module.DIDNotFound):
        await resolver.resolve(profile, TEST_DP3)


@pytest.mark.asyncio
async def test_record_removal(
    profile: Profile,
    resolver: PeerDID3Resolver,
    peer2_resolver: PeerDID2Resolver,
):
    """Test resolver setup."""
    await peer2_resolver.resolve(profile, TEST_DP2)
    assert await resolver.resolve(profile, TEST_DP3)
    record = ConnRecord(
        connection_id="test",
        my_did=TEST_DP2,
        their_did=TEST_DP3,
    )
    record.state = ConnRecord.STATE_DELETED
    async with profile.session() as session:
        await record.emit_event(session, record.serialize())

    with pytest.raises(test_module.DIDNotFound):
        await resolver.resolve(profile, TEST_DP3)



"""Classes for managing a revocation registry."""

import hashlib
import http
import logging
import os
import re
from os.path import join
from pathlib import Path
from typing import Optional

import base58
from requests import Session
from requests.exceptions import RequestException

from ...indy.util import indy_client_dir
from ..error import RevocationError

LOGGER = logging.getLogger(__name__)


class RevocationRegistry:
    """Manage a revocation registry and tails file."""

    MIN_SIZE = 4
    MAX_SIZE = 32768

    def __init__(
        self,
        registry_id: Optional[str] = None,
        *,
        cred_def_id: Optional[str] = None,
        issuer_did: Optional[str] = None,
        max_creds: Optional[int] = None,
        reg_def_type: Optional[str] = None,
        tag: Optional[str] = None,
        tails_local_path: Optional[str] = None,
        tails_public_uri: Optional[str] = None,
        tails_hash: Optional[str] = None,
        reg_def: Optional[dict] = None,
    ):
        """Initialize the revocation registry instance."""
        self._cred_def_id = cred_def_id
        self._issuer_did = issuer_did
        self._max_creds = max_creds
        self._reg_def_type = reg_def_type
        self._registry_id = registry_id
        self._tag = tag
        self._tails_local_path = tails_local_path
        self._tails_public_uri = tails_public_uri
        self._tails_hash = tails_hash
        self._reg_def = reg_def

    @classmethod
    def from_definition(
        cls, revoc_reg_def: dict, public_def: bool
    ) -> "RevocationRegistry":
        """Initialize a revocation registry instance from a definition."""
        rev_reg = None

        reg_id = revoc_reg_def["id"]
        tails_location = revoc_reg_def["value"]["tailsLocation"]
        issuer_did_match = re.match(r"^.*?([^:]*):3:CL:.*", revoc_reg_def["credDefId"])
        issuer_did = issuer_did_match.group(1) if issuer_did_match else None
        init = {
            "cred_def_id": revoc_reg_def["credDefId"],
            "issuer_did": issuer_did,
            "reg_def_type": revoc_reg_def["revocDefType"],
            "max_creds": revoc_reg_def["value"]["maxCredNum"],
            "tag": revoc_reg_def["tag"],
            "tails_hash": revoc_reg_def["value"]["tailsHash"],
            "reg_def": revoc_reg_def,
        }
        if public_def:
            init["tails_public_uri"] = tails_location
            rev_reg = cls(reg_id, **init)  # ignores def ver, issuance type, public keys
            rev_reg.tails_local_path = rev_reg.get_receiving_tails_local_path()
        else:
            init["tails_local_path"] = tails_location
            rev_reg = cls(reg_id, **init)  # ignores def ver, issuance type, public keys

        return rev_reg

    @property
    def cred_def_id(self) -> str:
        """Accessor for the credential definition ID."""
        return self._cred_def_id

    @property
    def issuer_did(self) -> str:
        """Accessor for the issuer DID."""
        return self._issuer_did

    @property
    def max_creds(self) -> int:
        """Accessor for the maximum number of issued credentials."""
        return self._max_creds

    @property
    def reg_def_type(self) -> str:
        """Accessor for the revocation registry type."""
        return self._reg_def_type

    @property
    def reg_def(self) -> dict:
        """Accessor for the revocation registry definition."""
        return self._reg_def

    @property
    def registry_id(self) -> str:
        """Accessor for the revocation registry ID."""
        return self._registry_id

    @property
    def tag(self) -> str:
        """Accessor for the tag part of the revoc. reg. ID."""
        return self._tag

    @property
    def tails_hash(self) -> str:
        """Accessor for the tails file hash."""
        return self._tails_hash

    @property
    def tails_local_path(self) -> str:
        """Accessor for the tails file local path."""
        return self._tails_local_path

    @tails_local_path.setter
    def tails_local_path(self, new_path: str):
        """Setter for the tails file local path."""
        self._tails_local_path = new_path

    @property
    def tails_public_uri(self) -> str:
        """Accessor for the tails file public URI."""
        return self._tails_public_uri

    @tails_public_uri.setter
    def tails_public_uri(self, new_uri: str):
        """Setter for the tails file public URI."""
        self._tails_public_uri = new_uri

    def get_receiving_tails_local_path(self):
        """Make the local path to the tails file we download from remote URI."""
        if self._tails_local_path:
            return self._tails_local_path

        tails_dir = indy_client_dir(join("tails", self.registry_id), create=False)
        return join(tails_dir, self._tails_hash)

    def has_local_tails_file(self) -> bool:
        """Test if the tails file exists locally."""
        tails_file_path = Path(self.get_receiving_tails_local_path())
        return tails_file_path.is_file()

    async def retrieve_tails(self):
        """Fetch the tails file from the public URI."""
        if not self._tails_public_uri:
            raise RevocationError("Tails file public URI is empty")

        LOGGER.info(
            "Downloading the tails file for the revocation registry: %s",
            self.registry_id,
        )

        tails_file_path = Path(self.get_receiving_tails_local_path())
        tails_file_dir = tails_file_path.parent
        if not tails_file_dir.exists():
            tails_file_dir.mkdir(parents=True)

        buffer_size = 65536  # should be multiple of 32 bytes for sha256
        file_hasher = hashlib.sha256()
        with open(tails_file_path, "wb", buffer_size) as tails_file:
            with Session() as req_session:
                try:
                    resp = req_session.get(self._tails_public_uri, stream=True)
                    # Should this directly raise an Error?
                    if resp.status_code != http.HTTPStatus.OK:
                        LOGGER.warning(
                            f"Unexpected status code for tails file: {resp.status_code}"
                        )
                    for buf in resp.iter_content(chunk_size=buffer_size):
                        tails_file.write(buf)
                        file_hasher.update(buf)
                except RequestException as rx:
                    raise RevocationError(f"Error retrieving tails file: {rx}")

        download_tails_hash = base58.b58encode(file_hasher.digest()).decode("utf-8")
        if download_tails_hash != self.tails_hash:
            try:
                os.remove(tails_file_path)
                tails_file_dir.rmdir()
            except OSError as err:
                LOGGER.warning(f"Could not delete invalid tails file: {err}")

            raise RevocationError("The hash of the downloaded tails file does not match.")

        self.tails_local_path = str(tails_file_path)
        return self.tails_local_path

    async def get_or_fetch_local_tails_path(self):
        """Get the local tails path, retrieving from the remote if necessary."""
        tails_file_path = self.get_receiving_tails_local_path()
        if Path(tails_file_path).is_file():
            return tails_file_path
        return await self.retrieve_tails()

    def __repr__(self) -> str:
        """Return a human readable representation of this class."""
        items = ("{}={}".format(k, repr(v)) for k, v in self.__dict__.items())
        return "<{}({})>".format(self.__class__.__name__, ", ".join(items))



"""Utilities for specifying which verification method is in use for a given DID."""

import logging
from abc import ABC, abstractmethod
from typing import Literal, Optional

from pydid import DIDDocument, VerificationMethod

from acapy_agent.wallet.key_type import BLS12381G2, ED25519, P256
from acapy_agent.wallet.keys.manager import (
    MultikeyManager,
    key_type_from_multikey,
    multikey_from_verification_method,
)

from ..core.error import BaseError
from ..core.profile import Profile
from ..did.did_key import DIDKey
from ..resolver.did_resolver import DIDResolver

LOGGER = logging.getLogger(__name__)


ProofPurposeStr = Literal[
    "assertionMethod",
    "authentication",
    "capabilityDelegation",
    "capabilityInvocation",
]
PROOF_PURPOSES = (
    "authentication",
    "assertionMethod",
    "capabilityInvocation",
    "capabilityDelegation",
)


class VerificationKeyStrategyError(BaseError):
    """Raised on issues with verfication method derivation."""


class BaseVerificationKeyStrategy(ABC):
    """Base class for defining which verification method is in use."""

    @abstractmethod
    async def get_verification_method_id_for_did(
        self,
        did: str,
        profile: Profile,
        *,
        proof_type: Optional[str] = None,
        proof_purpose: Optional[ProofPurposeStr] = None,
        verification_method_id: Optional[str] = None,
    ) -> str:
        """Find suitable VerificationMethod.

        Given a DID and other verification method requirements,
        find and return the first suitable verification method ID.
        Throws if no suitable verification method

        :params did: the did
        :params profile: context of the call
        :params proof_type: the JSON-LD proof type which the verification method
            should be able to produce.
        :params proof_purpose: the verkey relationship (assertionMethod, keyAgreement, ..)
        :params verification_method_id: the verification method ID which must match.
        :returns str: the first suitable verification method
        """
        ...


class DefaultVerificationKeyStrategy(BaseVerificationKeyStrategy):
    """A basic implementation for verkey strategy."""

    def __init__(self):
        """Initialize the key types mapping.

        Map of LDP signature suite (proofType) to suitable key types
        """
        self.key_types_mapping = {
            "Ed25519Signature2018": [ED25519],
            "Ed25519Signature2020": [ED25519],
            "EcdsaSecp256r1Signature2019": [P256],
            "BbsBlsSignature2020": [BLS12381G2],
            "BbsBlsSignatureProof2020": [BLS12381G2],
        }

    async def get_verification_method_id_for_did(
        self,
        did: str,
        profile: Profile,
        *,
        proof_type: Optional[str] = None,
        proof_purpose: Optional[ProofPurposeStr] = None,
        verification_method_id: Optional[str] = None,
    ) -> str:
        """Find suitable VerificationMethod."""
        proof_purpose = proof_purpose or "assertionMethod"

        if proof_purpose not in PROOF_PURPOSES:
            raise ValueError("Invalid proof purpose")

        if proof_type is not None:
            suitable_key_types = self.key_types_mapping.get(proof_type)
        else:
            # any key is suitable if no proof type requirement set
            suitable_key_types = list(
                {val for values in self.key_types_mapping.values() for val in values}
            )
        if not suitable_key_types:
            raise VerificationKeyStrategyError(
                f"proof type {proof_type} is not supported"
            )

        # handle default hardcoded cases
        if did.startswith("did:key:"):
            didkey = DIDKey.from_did(did)
            vm_id = didkey.key_id
            if didkey.key_type not in suitable_key_types:
                raise VerificationKeyStrategyError(
                    f"DID {did} has wrong key type for proof type {proof_type}"
                )
            if verification_method_id is not None and vm_id != verification_method_id:
                raise VerificationKeyStrategyError(
                    f"Verification method ID {verification_method_id} \
                        cannot be used with DID {did}"
                )
            return vm_id
        elif did.startswith("did:sov:"):
            # key-1 is what uniresolver uses for key id
            vm_id = did + "#key-1"
            if ED25519 not in suitable_key_types:
                raise VerificationKeyStrategyError(
                    "did:sov only capable of ED25519 based proof types"
                )
            if verification_method_id is not None and vm_id != verification_method_id:
                raise VerificationKeyStrategyError(
                    f"Verification method ID {verification_method_id} \
                        cannot be used with DID {did}"
                )
            return vm_id

        # else, handle generically for any DID

        # TODO - if the local representation of the DID contains all this information,
        #   DID resolution cost could be avoided. However, for now there is not adequate
        #   information locally to determine if a DID/VM is suitable.

        # shortcut path: if a VM ID is specified, fetch it with multikey and perform
        # basic checks of VM suitability.
        # NOTE: this skips the proofPurpose check, as that is not currently possible
        #   without resolving the DID (expensive)
        if verification_method_id is not None:
            async with profile.session() as session:
                key_manager = MultikeyManager(session=session)
                key_info = await key_manager.resolve_and_bind_kid(verification_method_id)
                key_type = key_type_from_multikey(multikey=key_info["multikey"])
                if key_type not in suitable_key_types:
                    raise VerificationKeyStrategyError(
                        f"VerificationMethod {verification_method_id} has wrong key type \
                        for proof type {proof_type}"
                    )
                return verification_method_id

        resolver = profile.inject(DIDResolver)
        doc_raw = await resolver.resolve(profile=profile, did=did)
        doc = DIDDocument.deserialize(doc_raw)

        # get verification methods for the proof purpose
        methods_or_refs = doc_raw.get(proof_purpose, [])

        # apply various filters to determine set of suitable verification methods
        suitable_methods = []
        async with profile.session() as session:
            key_manager = MultikeyManager(session=session)
            for method_or_ref in methods_or_refs:
                # Dereference any refs in the verification relationship
                if isinstance(method_or_ref, str):
                    vm_id = await resolver.dereference_verification_method(
                        profile, method_or_ref, document=doc
                    )
                else:
                    vm_id = VerificationMethod.deserialize(method_or_ref)

                vm_multikey = multikey_from_verification_method(vm_id)

                # filter methods by key type expected for proof_type
                vm_key_type = key_type_from_multikey(vm_multikey)
                if vm_key_type not in suitable_key_types:
                    continue

                # filter methods for keys actually owned by the wallet
                if not await key_manager.multikey_exists(
                    multikey_from_verification_method(vm_id)
                ):
                    continue

                # survived all filters
                suitable_methods.append(vm_id)

        if not suitable_methods:
            raise VerificationKeyStrategyError(
                f"No matching verification method found for did {did} with proof "
                f"type {proof_type} and purpose {proof_purpose}"
            )

        if len(suitable_methods) > 1:
            LOGGER.info(
                (
                    "More than 1 verification method matched for did %s with proof "
                    "type %s and purpose %s; returning the first: %s"
                ),
                did,
                proof_type,
                proof_purpose,
                suitable_methods[0].id,
            )

        return suitable_methods[0].id



"""acapy_agent package entry point."""

import logging
import os
import sys

LOGGER = logging.getLogger(__name__)


def init_debug(args):
    """Initialize debugging environment."""

    ENABLE_PTVSD = os.getenv("ENABLE_PTVSD", "").lower()
    ENABLE_PTVSD = ENABLE_PTVSD and ENABLE_PTVSD not in ("false", "0")

    ENABLE_PYDEVD_PYCHARM = os.getenv("ENABLE_PYDEVD_PYCHARM", "").lower()
    ENABLE_PYDEVD_PYCHARM = ENABLE_PYDEVD_PYCHARM and ENABLE_PYDEVD_PYCHARM not in (
        "false",
        "0",
    )
    PYDEVD_PYCHARM_HOST = os.getenv("PYDEVD_PYCHARM_HOST", "localhost")
    PYDEVD_PYCHARM_AGENT_PORT = int(os.getenv("PYDEVD_PYCHARM_AGENT_PORT", 5001))

    # --debug to use microsoft's visual studio remote debugger
    if ENABLE_PTVSD or "--debug" in args:
        DAP_HOST = os.getenv("PTVSD_HOST", None) or os.getenv("DAP_HOST", "localhost")
        DAP_PORT = int(os.getenv("PTVSD_PORT", None) or os.getenv("DAP_PORT", 5678))
        try:
            import debugpy

            debugpy.listen((DAP_HOST, DAP_PORT))
            LOGGER.info(
                f"=== Waiting for debugger to attach to {DAP_HOST}:{DAP_PORT} ==="
            )
            debugpy.wait_for_client()
        except ImportError:
            LOGGER.error("debugpy library was not found")

    if ENABLE_PYDEVD_PYCHARM or "--debug-pycharm" in args:
        try:
            import pydevd_pycharm

            LOGGER.info(
                "aca-py remote debugging to "
                f"{PYDEVD_PYCHARM_HOST}:{PYDEVD_PYCHARM_AGENT_PORT}"
            )
            pydevd_pycharm.settrace(
                host=PYDEVD_PYCHARM_HOST,
                port=PYDEVD_PYCHARM_AGENT_PORT,
                stdoutToServer=True,
                stderrToServer=True,
                suspend=False,
            )
        except ImportError:
            LOGGER.error("pydevd_pycharm library was not found")


def run(args):
    """Execute aca-py."""
    from .commands import run_command  # noqa

    if len(args) > 1 and args[1] and args[1][0] != "-":
        command = args[1]
        args = args[2:]
    else:
        command = None
        args = args[1:]

    run_command(command, args)


def script_main():
    """Run the main function as a script for poetry."""
    main(sys.argv)


def main(args):
    """Execute default entry point."""
    init_debug(args)
    run(args)


if __name__ == "__main__":
    script_main()



from unittest import IsolatedAsyncioTestCase

from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......tests import mock
from ......transport.inbound.receipt import MessageReceipt
from ......utils.testing import create_test_profile
from ...handlers import refused_transaction_response_handler as test_module
from ...messages.refused_transaction_response import RefusedTransactionResponse


class TestRefusedTransactionResponseHandler(IsolatedAsyncioTestCase):
    async def test_called(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "TransactionManager", autospec=True
        ) as mock_tran_mgr:
            mock_tran_mgr.return_value.receive_refuse_response = mock.CoroutineMock()
            request_context.message = RefusedTransactionResponse()
            request_context.connection_ready = True
            handler = test_module.RefusedTransactionResponseHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_tran_mgr.return_value.receive_refuse_response.assert_called_once_with(
            request_context.message
        )
        assert not responder.messages

    async def test_called_not_ready(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "TransactionManager", autospec=True
        ) as mock_tran_mgr:
            mock_tran_mgr.return_value.receive_refuse_response = mock.CoroutineMock()
            request_context.message = RefusedTransactionResponse()
            request_context.connection_ready = False
            handler = test_module.RefusedTransactionResponseHandler()
            responder = MockResponder()
            with self.assertRaises(test_module.HandlerException):
                await handler.handle(request_context, responder)

            assert not responder.messages

    async def test_called_x(self):
        request_context = RequestContext.test_context(await create_test_profile())
        request_context.message_receipt = MessageReceipt()
        request_context.connection_record = mock.MagicMock()

        with mock.patch.object(
            test_module, "TransactionManager", autospec=True
        ) as mock_tran_mgr:
            mock_tran_mgr.return_value.receive_refuse_response = mock.CoroutineMock(
                side_effect=test_module.TransactionManagerError()
            )
            request_context.message = RefusedTransactionResponse()
            request_context.connection_ready = True
            handler = test_module.RefusedTransactionResponseHandler()
            responder = MockResponder()
            await handler.handle(request_context, responder)

        mock_tran_mgr.return_value.receive_refuse_response.assert_called_once_with(
            request_context.message
        )
        assert not responder.messages



"""Utilities for dif presentation exchange attachment."""

import json
import logging
import time
from typing import Optional, Tuple, Union

from ....core.error import BaseError
from ....core.profile import Profile
from ....indy.holder import IndyHolder, IndyHolderError
from ....indy.models.xform import indy_proof_req2non_revoc_intervals
from ....ledger.multiple_ledger.ledger_requests_executor import (
    GET_REVOC_REG_DELTA,
    GET_SCHEMA,
    IndyLedgerRequestsExecutor,
)
from ....multitenant.base import BaseMultitenantManager
from ....revocation.models.revocation_registry import RevocationRegistry
from ..v1_0.models.presentation_exchange import V10PresentationExchange
from ..v2_0.messages.pres_format import V20PresFormat
from ..v2_0.models.pres_exchange import V20PresExRecord

LOGGER = logging.getLogger(__name__)


class IndyPresExchHandlerError(BaseError):
    """Base class for Indy Presentation Exchange related errors."""


class IndyPresExchHandler:
    """Base Presentation Exchange Handler."""

    def __init__(
        self,
        profile: Profile,
    ):
        """Initialize PresExchange Handler."""
        super().__init__()
        self._profile = profile

    async def return_presentation(
        self,
        pres_ex_record: Union[V10PresentationExchange, V20PresExRecord],
        requested_credentials: Optional[dict] = None,
    ) -> dict:
        """Return Indy proof request as dict."""
        # Get all credentials for this presentation
        holder = self._profile.inject(IndyHolder)
        credentials = {}
        requested_credentials = requested_credentials or {}

        # extract credential ids and non_revoked
        requested_referents = {}
        if isinstance(pres_ex_record, V20PresExRecord):
            proof_request = pres_ex_record.pres_request.attachment(
                V20PresFormat.Format.INDY
            )
            # If indy filter fails try anoncreds filter format. This is for a
            # non-anoncreds agent that gets a anoncreds format proof request and
            # should removed when indy format is fully retired.
            if not proof_request:
                proof_request = pres_ex_record.pres_request.attachment(
                    V20PresFormat.Format.ANONCREDS
                )
        elif isinstance(pres_ex_record, V10PresentationExchange):
            proof_request = pres_ex_record._presentation_request.ser
        non_revoc_intervals = indy_proof_req2non_revoc_intervals(proof_request)
        attr_creds = requested_credentials.get("requested_attributes", {})
        req_attrs = proof_request.get("requested_attributes", {})
        for reft in attr_creds:
            requested_referents[reft] = {"cred_id": attr_creds[reft]["cred_id"]}
            if reft in req_attrs and reft in non_revoc_intervals:
                requested_referents[reft]["non_revoked"] = non_revoc_intervals[reft]
        pred_creds = requested_credentials.get("requested_predicates", {})
        req_preds = proof_request.get("requested_predicates", {})
        for reft in pred_creds:
            requested_referents[reft] = {"cred_id": pred_creds[reft]["cred_id"]}
            if reft in req_preds and reft in non_revoc_intervals:
                requested_referents[reft]["non_revoked"] = non_revoc_intervals[reft]
        # extract mapping of presentation referents to credential ids
        for reft in requested_referents:
            credential_id = requested_referents[reft]["cred_id"]
            if credential_id not in credentials:
                credentials[credential_id] = json.loads(
                    await holder.get_credential(credential_id)
                )
        # remove any timestamps that cannot correspond to non-revoc intervals
        for r in ("requested_attributes", "requested_predicates"):
            for reft, req_item in requested_credentials.get(r, {}).items():
                if not credentials[req_item["cred_id"]].get(
                    "rev_reg_id"
                ) and req_item.pop("timestamp", None):
                    LOGGER.info(
                        f"Removed superfluous timestamp from requested_credentials {r} "
                        f"{reft} for non-revocable credential {req_item['cred_id']}"
                    )
        # Get all schemas, credential definitions, and revocation registries in use
        schemas = {}
        cred_defs = {}
        revocation_registries = {}

        for credential in credentials.values():
            schema_id = credential["schema_id"]
            multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(self._profile)
            else:
                ledger_exec_inst = self._profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    schema_id,
                    txn_record_type=GET_SCHEMA,
                )
            )[1]
            async with ledger:
                if schema_id not in schemas:
                    schemas[schema_id] = await ledger.get_schema(schema_id)
                cred_def_id = credential["cred_def_id"]
                if cred_def_id not in cred_defs:
                    cred_defs[cred_def_id] = await ledger.get_credential_definition(
                        cred_def_id
                    )
                if credential.get("rev_reg_id"):
                    revocation_registry_id = credential["rev_reg_id"]
                    if revocation_registry_id not in revocation_registries:
                        revocation_registries[revocation_registry_id] = (
                            RevocationRegistry.from_definition(
                                await ledger.get_revoc_reg_def(revocation_registry_id),
                                True,
                            )
                        )
        # Get delta with non-revocation interval defined in "non_revoked"
        # of the presentation request or attributes
        epoch_now = int(time.time())
        revoc_reg_deltas = {}
        for precis in requested_referents.values():  # cred_id, non-revoc interval
            credential_id = precis["cred_id"]
            if not credentials[credential_id].get("rev_reg_id"):
                continue
            if "timestamp" in precis:
                continue
            rev_reg_id = credentials[credential_id]["rev_reg_id"]
            multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(self._profile)
            else:
                ledger_exec_inst = self._profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    rev_reg_id,
                    txn_record_type=GET_REVOC_REG_DELTA,
                )
            )[1]
            async with ledger:
                reft_non_revoc_interval = precis.get("non_revoked")
                if reft_non_revoc_interval:
                    key = (
                        f"{rev_reg_id}_"
                        f"{reft_non_revoc_interval.get('from', 0)}_"
                        f"{reft_non_revoc_interval.get('to', epoch_now)}"
                    )
                    if key not in revoc_reg_deltas:
                        (delta, delta_timestamp) = await ledger.get_revoc_reg_delta(
                            rev_reg_id,
                            reft_non_revoc_interval.get("from", 0),
                            reft_non_revoc_interval.get("to", epoch_now),
                        )
                        revoc_reg_deltas[key] = (
                            rev_reg_id,
                            credential_id,
                            delta,
                            delta_timestamp,
                        )
                    for stamp_me in requested_referents.values():
                        # often one cred satisfies many requested attrs/preds
                        if stamp_me["cred_id"] == credential_id:
                            stamp_me["timestamp"] = revoc_reg_deltas[key][3]
        # Get revocation states to prove non-revoked
        revocation_states = {}
        for (
            rev_reg_id,
            credential_id,
            delta,
            delta_timestamp,
        ) in revoc_reg_deltas.values():
            if rev_reg_id not in revocation_states:
                revocation_states[rev_reg_id] = {}
            rev_reg = revocation_registries[rev_reg_id]
            tails_local_path = await rev_reg.get_or_fetch_local_tails_path()
            try:
                revocation_states[rev_reg_id][delta_timestamp] = json.loads(
                    await holder.create_revocation_state(
                        credentials[credential_id]["cred_rev_id"],
                        rev_reg.reg_def,
                        delta,
                        delta_timestamp,
                        tails_local_path,
                    )
                )
            except IndyHolderError as e:
                LOGGER.error(
                    f"Failed to create revocation state: {e.error_code}, {e.message}"
                )
                raise e
        for referent, precis in requested_referents.items():
            if "timestamp" not in precis:
                continue
            if referent in requested_credentials["requested_attributes"]:
                requested_credentials["requested_attributes"][referent]["timestamp"] = (
                    precis["timestamp"]
                )
            if referent in requested_credentials["requested_predicates"]:
                requested_credentials["requested_predicates"][referent]["timestamp"] = (
                    precis["timestamp"]
                )
        indy_proof_json = await holder.create_presentation(
            proof_request,
            requested_credentials,
            schemas,
            cred_defs,
            revocation_states,
        )
        indy_proof = json.loads(indy_proof_json)
        return indy_proof

    async def process_pres_identifiers(
        self,
        identifiers: list,
    ) -> Tuple[dict, dict, dict, dict]:
        """Return schemas, cred_defs, rev_reg_defs, rev_reg_entries."""
        schema_ids = []
        cred_def_ids = []

        schemas = {}
        cred_defs = {}
        rev_reg_defs = {}
        rev_reg_entries = {}

        for identifier in identifiers:
            schema_ids.append(identifier["schema_id"])
            cred_def_ids.append(identifier["cred_def_id"])
            multitenant_mgr = self._profile.inject_or(BaseMultitenantManager)
            if multitenant_mgr:
                ledger_exec_inst = IndyLedgerRequestsExecutor(self._profile)
            else:
                ledger_exec_inst = self._profile.inject(IndyLedgerRequestsExecutor)
            ledger = (
                await ledger_exec_inst.get_ledger_for_identifier(
                    identifier["schema_id"],
                    txn_record_type=GET_SCHEMA,
                )
            )[1]
            async with ledger:
                # Build schemas for anoncreds
                if identifier["schema_id"] not in schemas:
                    schemas[identifier["schema_id"]] = await ledger.get_schema(
                        identifier["schema_id"]
                    )

                if identifier["cred_def_id"] not in cred_defs:
                    cred_defs[
                        identifier["cred_def_id"]
                    ] = await ledger.get_credential_definition(identifier["cred_def_id"])

                if identifier.get("rev_reg_id"):
                    if identifier["rev_reg_id"] not in rev_reg_defs:
                        rev_reg_defs[
                            identifier["rev_reg_id"]
                        ] = await ledger.get_revoc_reg_def(identifier["rev_reg_id"])

                    if identifier.get("timestamp"):
                        rev_reg_entries.setdefault(identifier["rev_reg_id"], {})

                        if (
                            identifier["timestamp"]
                            not in rev_reg_entries[identifier["rev_reg_id"]]
                        ):
                            (
                                found_rev_reg_entry,
                                _found_timestamp,
                            ) = await ledger.get_revoc_reg_entry(
                                identifier["rev_reg_id"], identifier["timestamp"]
                            )
                            rev_reg_entries[identifier["rev_reg_id"]][
                                identifier["timestamp"]
                            ] = found_rev_reg_entry
        return (
            schemas,
            cred_defs,
            rev_reg_defs,
            rev_reg_entries,
        )



"""Test the InvitationCreator class."""

from unittest.mock import MagicMock

import pytest

from ..manager import InvitationCreator, OutOfBandManagerError


@pytest.mark.parametrize(
    "args",
    [
        ({}),
        ({"metadata": "test"}),
        ({"attachments": "test", "multi_use": True}),
        ({"hs_protos": "test", "public": True, "use_did": True}),
        ({"hs_protos": "test", "public": True, "use_did_method": True}),
        ({"hs_protos": "test", "use_did": True, "use_did_method": True}),
        ({"hs_protos": "test", "create_unique_did": True}),
        ({"hs_protos": "test", "use_did_method": "some_did_method"}),
    ],
)
def test_init_param_checking_x(args):
    with pytest.raises(OutOfBandManagerError):
        InvitationCreator(
            profile=MagicMock(), route_manager=MagicMock(), oob=MagicMock(), **args
        )



"""Multitenant admin routes."""

from aiohttp import web
from aiohttp_apispec import (
    docs,
    match_info_schema,
    querystring_schema,
    request_schema,
    response_schema,
)
from marshmallow import ValidationError, fields, validate, validates_schema

from ...admin.decorators.auth import admin_authentication
from ...admin.request_context import AdminRequestContext
from ...core.error import BaseError
from ...core.profile import ProfileManagerProvider
from ...messaging.models.base import BaseModelError
from ...messaging.models.openapi import OpenAPISchema
from ...messaging.models.paginated_query import (
    PaginatedQuerySchema,
    get_paginated_query_params,
)
from ...messaging.valid import UUID4_EXAMPLE, JSONWebToken
from ...multitenant.base import BaseMultitenantManager
from ...storage.error import StorageError, StorageNotFoundError
from ...utils.endorsement_setup import attempt_auto_author_with_endorser_setup
from ...wallet.error import WalletSettingsError
from ...wallet.models.wallet_record import WalletRecord, WalletRecordSchema
from ..error import WalletKeyMissingError

ACAPY_LIFECYCLE_CONFIG_FLAG_MAP = {
    "ACAPY_AUTO_ACCEPT_INVITES": "debug.auto_accept_invites",
    "ACAPY_AUTO_ACCEPT_REQUESTS": "debug.auto_accept_requests",
    "ACAPY_AUTO_PING_CONNECTION": "auto_ping_connection",
    "ACAPY_AUTO_PROMOTE_AUTHOR_DID": "endorser.auto_promote_author_did",
    "ACAPY_AUTO_REQUEST_ENDORSEMENT": "endorser.auto_request",
    "ACAPY_AUTO_RESPOND_CREDENTIAL_OFFER": "debug.auto_respond_credential_offer",
    "ACAPY_AUTO_RESPOND_CREDENTIAL_REQUEST": "debug.auto_respond_credential_request",
    "ACAPY_AUTO_RESPOND_MESSAGES": "debug.auto_respond_messages",
    "ACAPY_AUTO_VERIFY_PRESENTATION": "debug.auto_verify_presentation",
    "ACAPY_AUTO_WRITE_TRANSACTIONS": "endorser.auto_write",
    "ACAPY_CREATE_REVOCATION_TRANSACTIONS": "endorser.auto_create_rev_reg",
    "ACAPY_ENDORSER_ALIAS": "endorser.endorser_alias",
    "ACAPY_ENDORSER_INVITATION": "endorser.endorser_invitation",
    "ACAPY_ENDORSER_PUBLIC_DID": "endorser.endorser_public_did",
    "ACAPY_ENDORSER_ROLE": "endorser.protocol_role",
    "ACAPY_INVITE_PUBLIC": "debug.invite_public",
    "ACAPY_LOG_LEVEL": "log.level",
    "ACAPY_MONITOR_PING": "debug.monitor_ping",
    "ACAPY_NOTIFY_REVOCATION": "revocation.notify",
    "ACAPY_PRESERVE_EXCHANGE_RECORDS": "preserve_exchange_records",
    "ACAPY_PUBLIC_INVITES": "public_invites",
    "ACAPY_REQUESTS_THROUGH_PUBLIC_DID": "requests_through_public_did",
}

ACAPY_LIFECYCLE_CONFIG_FLAG_ARGS_MAP = {
    "auto-accept-invites": "debug.auto_accept_invites",
    "auto-accept-requests": "debug.auto_accept_requests",
    "auto-create-revocation-transactions": "endorser.auto_create_rev_reg",
    "auto-ping-connection": "auto_ping_connection",
    "auto-promote-author-did": "endorser.auto_promote_author_did",
    "auto-request-endorsement": "endorser.auto_request",
    "auto-respond-credential-offer": "debug.auto_respond_credential_offer",
    "auto-respond-credential-request": "debug.auto_respond_credential_request",
    "auto-respond-messages": "debug.auto_respond_messages",
    "auto-verify-presentation": "debug.auto_verify_presentation",
    "auto-write-transactions": "endorser.auto_write",
    "endorser-alias": "endorser.endorser_alias",
    "endorser-invitation": "endorser.endorser_invitation",
    "endorser-protocol-role": "endorser.protocol_role",
    "endorser-public-did": "endorser.endorser_public_did",
    "invite-public": "debug.invite_public",
    "log-level": "log.level",
    "monitor-ping": "debug.monitor_ping",
    "notify-revocation": "revocation.notify",
    "preserve-exchange-records": "preserve_exchange_records",
    "public-invites": "public_invites",
    "requests-through-public-did": "requests_through_public_did",
}

ACAPY_ENDORSER_FLAGS_DEPENDENT_ON_AUTHOR_ROLE = [
    "ACAPY_AUTO_REQUEST_ENDORSEMENT",
    "ACAPY_AUTO_WRITE_TRANSACTIONS",
    "ACAPY_CREATE_REVOCATION_TRANSACTIONS",
    "auto-request-endorsement",
    "auto-write-transactions",
    "auto-create-revocation-transactions",
]


def format_wallet_record(wallet_record: WalletRecord):
    """Serialize a WalletRecord object."""

    wallet_info = wallet_record.serialize()

    # Hide wallet wallet key
    if "wallet.key" in wallet_info["settings"]:
        del wallet_info["settings"]["wallet.key"]

    return wallet_info


def get_extra_settings_dict_per_tenant(tenant_settings: dict) -> dict:
    """Get per tenant settings to be applied when creating wallet."""

    endorser_role_flag = tenant_settings.get(
        "ACAPY_ENDORSER_ROLE"
    ) or tenant_settings.get("endorser-protocol-role")
    extra_settings = {}
    if endorser_role_flag and endorser_role_flag == "author":
        extra_settings["endorser.author"] = True
    elif endorser_role_flag and endorser_role_flag == "endorser":
        extra_settings["endorser.endorser"] = True
    for flag in tenant_settings.keys():
        if (
            flag in ACAPY_ENDORSER_FLAGS_DEPENDENT_ON_AUTHOR_ROLE
            and endorser_role_flag != "author"
        ):
            # These flags require endorser role as author, if not set as author then
            # this setting will be ignored.
            continue
        if flag != "ACAPY_ENDORSER_ROLE" and flag != "endorser-protocol-role":
            map_flag = ACAPY_LIFECYCLE_CONFIG_FLAG_MAP.get(
                flag
            ) or ACAPY_LIFECYCLE_CONFIG_FLAG_ARGS_MAP.get(flag)
            if not map_flag:
                continue
            extra_settings[map_flag] = tenant_settings[flag]
    return extra_settings


class MultitenantModuleResponseSchema(OpenAPISchema):
    """Response schema for multitenant module."""


class WalletIdMatchInfoSchema(OpenAPISchema):
    """Path parameters and validators for request taking wallet id."""

    wallet_id = fields.Str(
        required=True,
        metadata={"description": "Subwallet identifier", "example": UUID4_EXAMPLE},
    )


class CreateWalletRequestSchema(OpenAPISchema):
    """Request schema for adding a new wallet which will be registered by the agent."""

    wallet_name = fields.Str(
        metadata={"description": "Wallet name", "example": "MyNewWallet"}
    )

    wallet_key = fields.Str(
        metadata={
            "description": "Master key used for key derivation.",
            "example": "MySecretKey123",
        }
    )

    extra_settings = fields.Dict(
        required=False, metadata={"description": "Agent config key-value pairs"}
    )

    wallet_key_derivation = fields.Str(
        required=False,
        validate=validate.OneOf(["ARGON2I_MOD", "ARGON2I_INT", "RAW"]),
        metadata={"description": "Key derivation", "example": "RAW"},
    )

    wallet_type = fields.Str(
        dump_default="in_memory",
        required=False,
        validate=validate.OneOf(list(ProfileManagerProvider.MANAGER_TYPES)),
        metadata={
            "description": "Type of the wallet to create. Must be same as base wallet.",
            "example": "askar",
        },
    )

    wallet_dispatch_type = fields.Str(
        dump_default="default",
        validate=validate.OneOf(["default", "both", "base"]),
        metadata={
            "description": (
                "Webhook target dispatch type for this wallet. "
                "default: Dispatch only to webhooks associated with this wallet. "
                "base: Dispatch only to webhooks associated with the base wallet. "
                "both: Dispatch to both webhook targets."
            ),
            "example": "default",
        },
    )

    wallet_webhook_urls = fields.List(
        fields.Str(
            metadata={
                "description": "Optional webhook URL to receive webhook messages",
                "example": "http://localhost:8022/webhooks",
            }
        ),
        required=False,
        metadata={"description": "List of Webhook URLs associated with this subwallet"},
    )

    label = fields.Str(
        metadata={
            "description": (
                "Label for this wallet. This label is publicized (self-attested)"
                " to other agents as part of forming a connection."
            ),
            "example": "Alice",
        }
    )

    image_url = fields.Str(
        metadata={
            "description": (
                "Image url for this wallet. This image url is publicized"
                " (self-attested) to other agents as part of forming a connection."
            ),
            "example": "https://aries.ca/images/sample.png",
        }
    )

    key_management_mode = fields.Str(
        dump_default=WalletRecord.MODE_MANAGED,
        # MTODO: add unmanaged mode once implemented
        validate=validate.OneOf((WalletRecord.MODE_MANAGED,)),
        metadata={
            "description": "Key management method to use for this wallet.",
            "example": WalletRecord.MODE_MANAGED,
        },
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        Raises:
            ValidationError: If any of the fields do not validate

        """

        if data.get("wallet_type") == "indy":
            for field in ("wallet_key", "wallet_name"):
                if field not in data:
                    raise ValidationError("Missing required field", field)


class UpdateWalletRequestSchema(OpenAPISchema):
    """Request schema for updating a existing wallet."""

    wallet_dispatch_type = fields.Str(
        dump_default="default",
        validate=validate.OneOf(["default", "both", "base"]),
        metadata={
            "description": (
                "Webhook target dispatch type for this wallet. "
                "default: Dispatch only to webhooks associated with this wallet. "
                "base: Dispatch only to webhooks associated with the base wallet. "
                "both: Dispatch to both webhook targets."
            ),
            "example": "default",
        },
    )
    extra_settings = fields.Dict(
        required=False, metadata={"description": "Agent config key-value pairs"}
    )
    wallet_webhook_urls = fields.List(
        fields.Str(
            metadata={
                "description": "Optional webhook URL to receive webhook messages",
                "example": "http://localhost:8022/webhooks",
            }
        ),
        required=False,
        metadata={"description": "List of Webhook URLs associated with this subwallet"},
    )
    label = fields.Str(
        metadata={
            "description": (
                "Label for this wallet. This label is publicized (self-attested)"
                " to other agents as part of forming a connection."
            ),
            "example": "Alice",
        }
    )
    image_url = fields.Str(
        metadata={
            "description": (
                "Image url for this wallet. This image url is publicized"
                " (self-attested) to other agents as part of forming a connection."
            ),
            "example": "https://aries.ca/images/sample.png",
        }
    )


class CreateWalletResponseSchema(WalletRecordSchema):
    """Response schema for creating a wallet."""

    token = fields.Str(
        metadata={
            "description": "Authorization token to authenticate wallet requests",
            "example": JSONWebToken.EXAMPLE,
        }
    )


class RemoveWalletRequestSchema(OpenAPISchema):
    """Request schema for removing a wallet."""

    wallet_key = fields.Str(
        metadata={
            "description": (
                "Master key used for key derivation. Only required for unmanaged wallets."
            ),
            "example": "MySecretKey123",
        }
    )


class CreateWalletTokenRequestSchema(OpenAPISchema):
    """Request schema for creating a wallet token."""

    wallet_key = fields.Str(
        metadata={
            "description": (
                "Master key used for key derivation. Only required for unmanaged wallets."
            ),
            "example": "MySecretKey123",
        }
    )


class CreateWalletTokenResponseSchema(OpenAPISchema):
    """Response schema for creating a wallet token."""

    token = fields.Str(
        metadata={
            "description": "Authorization token to authenticate wallet requests",
            "example": JSONWebToken.EXAMPLE,
        }
    )


class WalletListSchema(OpenAPISchema):
    """Result schema for wallet list."""

    results = fields.List(
        fields.Nested(WalletRecordSchema()),
        metadata={"description": "List of wallet records"},
    )


class WalletListQueryStringSchema(PaginatedQuerySchema):
    """Parameters and validators for wallet list request query string."""

    wallet_name = fields.Str(
        metadata={"description": "Wallet name", "example": "MyNewWallet"}
    )


@docs(tags=["multitenancy"], summary="Query subwallets")
@querystring_schema(WalletListQueryStringSchema())
@response_schema(WalletListSchema(), 200, description="")
@admin_authentication
async def wallets_list(request: web.BaseRequest):
    """Request handler for listing all internal subwallets.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    profile = context.profile

    query = {}
    wallet_name = request.query.get("wallet_name")
    if wallet_name:
        query["wallet_name"] = wallet_name

    limit, offset, order_by, descending = get_paginated_query_params(request)

    try:
        async with profile.session() as session:
            records = await WalletRecord.query(
                session,
                tag_filter=query,
                limit=limit,
                offset=offset,
                order_by=order_by,
                descending=descending,
            )
        results = [format_wallet_record(record) for record in records]
    except (StorageError, BaseModelError) as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response({"results": results})


@docs(tags=["multitenancy"], summary="Get a single subwallet")
@match_info_schema(WalletIdMatchInfoSchema())
@response_schema(WalletRecordSchema(), 200, description="")
@admin_authentication
async def wallet_get(request: web.BaseRequest):
    """Request handler for getting a single subwallet.

    Args:
        request: aiohttp request object

    Raises:
        HTTPNotFound: if wallet_id does not match any known wallets

    """

    context: AdminRequestContext = request["context"]
    profile = context.profile
    wallet_id = request.match_info["wallet_id"]

    try:
        async with profile.session() as session:
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)
        result = format_wallet_record(wallet_record)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except BaseModelError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Create a subwallet")
@request_schema(CreateWalletRequestSchema)
@response_schema(CreateWalletResponseSchema(), 200, description="")
@admin_authentication
async def wallet_create(request: web.BaseRequest):
    """Request handler for adding a new subwallet for handling by the agent.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    body = await request.json()

    base_wallet_type = context.profile.settings.get("wallet.type")
    sub_wallet_type = body.get("wallet_type", base_wallet_type)
    key_management_mode = body.get("key_management_mode") or WalletRecord.MODE_MANAGED
    wallet_key = body.get("wallet_key")
    wallet_webhook_urls = body.get("wallet_webhook_urls") or []
    wallet_dispatch_type = body.get("wallet_dispatch_type") or "default"
    extra_settings = body.get("extra_settings") or {}
    # If no webhooks specified, then dispatch only to base webhook targets
    if wallet_webhook_urls == []:
        wallet_dispatch_type = "base"

    settings = {
        "wallet.type": sub_wallet_type,
        "wallet.name": body.get("wallet_name"),
        "wallet.key": wallet_key,
        "wallet.webhook_urls": wallet_webhook_urls,
        "wallet.dispatch_type": wallet_dispatch_type,
    }
    extra_subwallet_setting = get_extra_settings_dict_per_tenant(extra_settings)
    settings.update(extra_subwallet_setting)

    label = body.get("label")
    image_url = body.get("image_url")
    key_derivation = body.get("wallet_key_derivation")
    if label:
        settings["default_label"] = label
    if image_url:
        settings["image_url"] = image_url
    if key_derivation:  # allow lower levels to handle default
        settings["wallet.key_derivation_method"] = key_derivation

    try:
        multitenant_mgr = context.profile.inject(BaseMultitenantManager)

        wallet_record = await multitenant_mgr.create_wallet(settings, key_management_mode)

        token = await multitenant_mgr.create_auth_token(wallet_record, wallet_key)

        wallet_profile = await multitenant_mgr.get_wallet_profile(
            context, wallet_record, extra_settings=settings
        )
        await attempt_auto_author_with_endorser_setup(wallet_profile)
    except BaseError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    result = {
        **format_wallet_record(wallet_record),
        "token": token,
    }
    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Update a subwallet")
@match_info_schema(WalletIdMatchInfoSchema())
@request_schema(UpdateWalletRequestSchema)
@response_schema(WalletRecordSchema(), 200, description="")
@admin_authentication
async def wallet_update(request: web.BaseRequest):
    """Request handler for updating a existing subwallet for handling by the agent.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]

    body = await request.json()
    wallet_webhook_urls = body.get("wallet_webhook_urls")
    wallet_dispatch_type = body.get("wallet_dispatch_type")
    label = body.get("label")
    image_url = body.get("image_url")
    extra_settings = body.get("extra_settings")

    if all(
        v is None
        for v in (
            wallet_webhook_urls,
            wallet_dispatch_type,
            label,
            image_url,
            extra_settings,
        )
    ):
        raise web.HTTPBadRequest(reason="At least one parameter is required.")

    # adjust wallet_dispatch_type according to wallet_webhook_urls
    if wallet_webhook_urls and wallet_dispatch_type is None:
        wallet_dispatch_type = "default"
    if wallet_webhook_urls == []:
        wallet_dispatch_type = "base"

    # only parameters that are not none are updated
    settings = {}
    if wallet_webhook_urls is not None:
        settings["wallet.webhook_urls"] = wallet_webhook_urls
    if wallet_dispatch_type is not None:
        settings["wallet.dispatch_type"] = wallet_dispatch_type
    if label is not None:
        settings["default_label"] = label
    if image_url is not None:
        settings["image_url"] = image_url
    extra_subwallet_setting = get_extra_settings_dict_per_tenant(extra_settings or {})
    settings.update(extra_subwallet_setting)

    try:
        multitenant_mgr = context.profile.inject(BaseMultitenantManager)
        wallet_record = await multitenant_mgr.update_wallet(wallet_id, settings)

        result = format_wallet_record(wallet_record)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except WalletSettingsError as err:
        raise web.HTTPBadRequest(reason=err.roll_up) from err

    return web.json_response(result)


@docs(tags=["multitenancy"], summary="Get auth token for a subwallet")
@request_schema(CreateWalletTokenRequestSchema)
@response_schema(CreateWalletTokenResponseSchema(), 200, description="")
@admin_authentication
async def wallet_create_token(request: web.BaseRequest):
    """Request handler for creating an authorization token for a specific subwallet.

    Args:
        request: aiohttp request object
    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]
    wallet_key = None

    if request.has_body:
        body = await request.json()
        wallet_key = body.get("wallet_key")

    profile = context.profile
    try:
        multitenant_mgr = profile.inject(BaseMultitenantManager)
        async with profile.session() as session:
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)

        if (not wallet_record.requires_external_key) and wallet_key:
            raise web.HTTPBadRequest(
                reason=(
                    f"Wallet {wallet_id} doesn't require the wallet key to be provided"
                )
            )

        token = await multitenant_mgr.create_auth_token(wallet_record, wallet_key)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except WalletKeyMissingError as err:
        raise web.HTTPUnauthorized(reason=err.roll_up) from err

    return web.json_response({"token": token})


@docs(
    tags=["multitenancy"],
    summary="Remove a subwallet",
)
@match_info_schema(WalletIdMatchInfoSchema())
@request_schema(RemoveWalletRequestSchema)
@response_schema(MultitenantModuleResponseSchema(), 200, description="")
@admin_authentication
async def wallet_remove(request: web.BaseRequest):
    """Request handler to remove a subwallet from agent and storage.

    Args:
        request: aiohttp request object.

    """

    context: AdminRequestContext = request["context"]
    wallet_id = request.match_info["wallet_id"]
    wallet_key = None

    if request.has_body:
        body = await request.json()
        wallet_key = body.get("wallet_key")

    profile = context.profile
    try:
        multitenant_mgr = profile.inject(BaseMultitenantManager)
        async with profile.session() as session:
            wallet_record = await WalletRecord.retrieve_by_id(session, wallet_id)

        if (not wallet_record.requires_external_key) and wallet_key:
            raise web.HTTPBadRequest(
                reason=(
                    f"Wallet {wallet_id} doesn't require the wallet key to be provided"
                )
            )

        await multitenant_mgr.remove_wallet(wallet_id, wallet_key)
    except StorageNotFoundError as err:
        raise web.HTTPNotFound(reason=err.roll_up) from err
    except WalletKeyMissingError as err:
        raise web.HTTPUnauthorized(reason=err.roll_up) from err

    return web.json_response({})


# MTODO: add wallet import route
# MTODO: add wallet export route
# MTODO: add rotate wallet key route


async def register(app: web.Application):
    """Register routes."""

    app.add_routes(
        [
            web.get("/multitenancy/wallets", wallets_list, allow_head=False),
            web.post("/multitenancy/wallet", wallet_create),
            web.get("/multitenancy/wallet/{wallet_id}", wallet_get, allow_head=False),
            web.put("/multitenancy/wallet/{wallet_id}", wallet_update),
            web.post("/multitenancy/wallet/{wallet_id}/token", wallet_create_token),
            web.post("/multitenancy/wallet/{wallet_id}/remove", wallet_remove),
        ]
    )


def post_process_routes(app: web.Application):
    """Amend swagger API."""

    # Add top-level tags description
    if "tags" not in app._state["swagger_dict"]:
        app._state["swagger_dict"]["tags"] = []
    app._state["swagger_dict"]["tags"].append(
        {"name": "multitenancy", "description": "Multitenant wallet management"}
    )



"""A problem report message."""

import logging
from enum import Enum

from marshmallow import EXCLUDE, ValidationError, validates_schema

from ....problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import CRED_20_PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.cred_problem_report_handler.CredProblemReportHandler"
)

LOGGER = logging.getLogger(__name__)


class ProblemReportReason(Enum):
    """Supported reason codes."""

    ISSUANCE_ABANDONED = "issuance-abandoned"
    RECORD_NOT_FOUND = "record-not-found"


class V20CredProblemReport(ProblemReport):
    """Class representing a problem report message."""

    class Meta:
        """Problem report metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "V20CredProblemReportSchema"
        message_type = CRED_20_PROBLEM_REPORT

    def __init__(self, *args, **kwargs):
        """Initialize problem report object."""
        super().__init__(*args, **kwargs)


class V20CredProblemReportSchema(ProblemReportSchema):
    """Problem report schema."""

    class Meta:
        """Schema metadata."""

        model_class = V20CredProblemReport
        unknown = EXCLUDE

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments

        """
        if not data.get("description", {}).get("code", ""):
            raise ValidationError("Value for description.code must be present")
        elif (
            data.get("description", {}).get("code", "")
            != ProblemReportReason.ISSUANCE_ABANDONED.value
        ):
            locales = list(data.get("description").keys())
            locales.remove("code")
            LOGGER.warning(
                "Unexpected error code received.\n"
                f"Code: {data.get('description').get('code')}, "
                f"Description: {data.get('description').get(locales[0])}"
            )



from unittest import IsolatedAsyncioTestCase

from ..ld_proof import V20CredExRecordLDProof


class TestV20CredExRecordLDProof(IsolatedAsyncioTestCase):
    async def test_record(self):
        same = [
            V20CredExRecordLDProof(cred_ex_ld_proof_id="dummy-0", cred_ex_id="abc")
        ] * 2
        diff = [
            V20CredExRecordLDProof(
                cred_ex_ld_proof_id="dummy-0",
                cred_ex_id="def",
            ),
            V20CredExRecordLDProof(
                cred_ex_ld_proof_id="dummy-0",
                cred_ex_id="abc",
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

        assert same[0].cred_ex_ld_proof_id == "dummy-0"



"""Represents a transaction request message."""

from typing import Optional

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.valid import UUID4_EXAMPLE
from ..message_types import PROTOCOL_PACKAGE, TRANSACTION_REQUEST

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.transaction_request_handler.TransactionRequestHandler"
)


class TransactionRequest(AgentMessage):
    """Class representing a transaction request message."""

    class Meta:
        """Metadata for a transaction request message."""

        handler_class = HANDLER_CLASS
        message_type = TRANSACTION_REQUEST
        schema_class = "TransactionRequestSchema"

    def __init__(
        self,
        *,
        transaction_id: Optional[str] = None,
        signature_request: Optional[dict] = None,
        timing: Optional[dict] = None,
        transaction_type: Optional[str] = None,
        messages_attach: Optional[dict] = None,
        endorser_write_txn: Optional[bool] = None,
        **kwargs,
    ):
        """Initialize the transaction request object.

        Args:
            transaction_id: The transaction id of the transaction record
            signature_request: The signature that is requested
            timing: The time till the endorser should endorse/refuse a transaction
            transaction_type: The type of transaction
            messages_attach: The attached message describing the actual transaction
            endorser_write_txn: Request Endorser to write the ledger transaction
            kwargs: Additional keyword arguments for the message
        """
        super().__init__(**kwargs)
        self.transaction_id = transaction_id
        self.signature_request = signature_request
        self.timing = timing
        self.transaction_type = transaction_type
        self.messages_attach = messages_attach
        self.endorser_write_txn = endorser_write_txn


class TransactionRequestSchema(AgentMessageSchema):
    """Transaction request schema class."""

    class Meta:
        """Transaction request schema metadata."""

        model_class = TransactionRequest
        unknown = EXCLUDE

    transaction_id = fields.Str(required=False, metadata={"example": UUID4_EXAMPLE})
    signature_request = fields.Dict(
        required=False,
        metadata={
            "example": {
                "context": "did:sov",
                "method": "add-signature",
                "signature_type": "<requested signature type>",
                "signer_goal_code": "transaction.endorse",
                "author_goal_code": "transaction.ledger.write",
            }
        },
    )
    timing = fields.Dict(
        required=False, metadata={"example": {"expires_time": "1597708800"}}
    )
    transaction_type = fields.Str(required=False, metadata={"example": "101"})
    messages_attach = fields.Dict(
        required=False,
        metadata={
            "example": {
                "@id": "143c458d-1b1c-40c7-ab85-4d16808ddf0a",
                "mime-type": "application/json",
                "data": {
                    "json": {
                        "endorser": "V4SGRU86Z58d6TV7PBUe6f",
                        "identifier": "LjgpST2rjsoxYegQDRm7EL",
                        "operation": {
                            "data": {
                                "attr_names": ["first_name", "last_name"],
                                "name": "test_schema",
                                "version": "2.1",
                            },
                            "type": "101",
                        },
                        "protocolVersion": 2,
                        "reqId": 1597766666168851000,
                        "signatures": {
                            "LjgpST2rjsox": (
                                "4uq1mUATKMn6Y9sTaGWyuPgjUEw5UBysWNbfSqCfnbm1Vnfw"
                            )
                        },
                        "taaAcceptance": {
                            "mechanism": "manual",
                            "taaDigest": (
                                "f50feca75664270842bd4202c2d6f23e4c6a7e0fc2feb9f62"
                            ),
                            "time": 1597708800,
                        },
                    }
                },
            }
        },
    )
    endorser_write_txn = fields.Boolean(
        required=False,
        metadata={
            "description": (
                "Request Endorser to write the ledger transaction, "
                "this parameter is deprecated and no longer supported."
            ),
            "example": False,
        },
    )



"""Abstract base class for linked data proofs."""

from abc import ABC
from typing import ClassVar, List, Optional, Union

from pyld import jsonld
from typing_extensions import TypedDict

from ..check import get_properties_without_context
from ..constants import SECURITY_CONTEXT_URL
from ..document_loader import DocumentLoaderMethod
from ..error import LinkedDataProofException
from ..purposes import _ProofPurpose as ProofPurpose
from ..validation_result import ProofResult


class DeriveProofResult(TypedDict):
    """Result dict for deriving a proof."""

    document: dict
    proof: Union[dict, List[dict]]


class LinkedDataProof(ABC):
    """Base Linked data proof."""

    signature_type: ClassVar[str]

    def __init__(
        self,
        *,
        proof: Optional[dict] = None,
        supported_derive_proof_types: Optional[List[str]] = None,
    ):
        """Initialize new LinkedDataProof instance."""
        self.proof = proof
        self.supported_derive_proof_types = supported_derive_proof_types

    async def create_proof(
        self,
        *,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> dict:
        """Create proof for document.

        Args:
            document (dict): The document to create the proof for
            purpose (ProofPurpose): The proof purpose to include in the proof
            document_loader (DocumentLoader): Document loader used for resolving

        Returns:
            dict: The proof object

        """
        raise LinkedDataProofException(
            f"{self.signature_type} signature suite does not support creating proofs"
        )

    async def verify_proof(
        self,
        *,
        proof: dict,
        document: dict,
        purpose: ProofPurpose,
        document_loader: DocumentLoaderMethod,
    ) -> ProofResult:
        """Verify proof against document and proof purpose.

        Args:
            proof (dict): The proof to verify
            document (dict): The document to verify the proof against
            purpose (ProofPurpose): The proof purpose to verify the proof against
            document_loader (DocumentLoader): Document loader used for resolving

        Returns:
            ValidationResult: The results of the proof verification

        """
        raise LinkedDataProofException(
            f"{self.signature_type} signature suite does not support verifying proofs"
        )

    async def derive_proof(
        self,
        *,
        proof: dict,
        document: dict,
        reveal_document: dict,
        document_loader: DocumentLoaderMethod,
        nonce: Optional[bytes] = None,
    ) -> DeriveProofResult:
        """Derive proof for document, returning derived document and proof.

        Args:
            proof (dict): The proof to derive from
            document (dict): The document to derive the proof for
            reveal_document (dict): The JSON-LD frame the revealed attributes
            document_loader (DocumentLoader): Document loader used for resolving
            nonce (bytes, optional): Nonce to use for the proof. Defaults to None.

        Returns:
            DeriveProofResult: The derived document and proof

        """
        raise LinkedDataProofException(
            f"{self.signature_type} signature suite does not support deriving proofs"
        )

    def _canonize(self, *, input, document_loader: DocumentLoaderMethod) -> str:
        """Canonize input document using URDNA2015 algorithm."""
        # application/n-quads format always returns str
        missing_properties = get_properties_without_context(input, document_loader)

        if len(missing_properties) > 0:
            raise LinkedDataProofException(
                f"{len(missing_properties)} attributes dropped. "
                f"Provide definitions in context to correct. {missing_properties}"
            )

        return jsonld.normalize(
            input,
            {
                "algorithm": "URDNA2015",
                "format": "application/n-quads",
                "documentLoader": document_loader,
            },
        )

    def _get_verification_method(
        self, *, proof: dict, document_loader: DocumentLoaderMethod
    ) -> dict:
        """Get verification method for proof."""

        verification_method = proof.get("verificationMethod")

        if isinstance(verification_method, dict):
            verification_method: str = verification_method.get("id")

        if not verification_method:
            raise LinkedDataProofException('No "verificationMethod" found in proof')

        # TODO: This should optionally use the context of the document?
        framed = jsonld.frame(
            verification_method,
            frame={
                "@context": SECURITY_CONTEXT_URL,
                "@embed": "@always",
                "id": verification_method,
            },
            options={
                "documentLoader": document_loader,
                "expandContext": SECURITY_CONTEXT_URL,
                # if we don't set base explicitly it will remove the base in returned
                # document (e.g. use key:z... instead of did:key:z...)
                # same as compactToRelative in jsonld.js
                "base": None,
            },
        )

        if not framed:
            raise LinkedDataProofException(
                f"Verification method {verification_method} not found"
            )

        if framed.get("revoked"):
            raise LinkedDataProofException("The verification method has been revoked.")

        return framed

    def match_proof(self, signature_type: str) -> bool:
        """Match signature type to signature type of this suite."""
        return signature_type == self.signature_type



import unittest

from aiohttp.web import HTTPUnprocessableEntity
from marshmallow.exceptions import ValidationError

from ...utils.extract_validation_error import extract_validation_error_message


class TestExtractValidationErrorMessage(unittest.TestCase):
    def test_validation_error_extracted(self):
        """Test that the validation error message is extracted when present."""
        validation_error = ValidationError({"field": ["Invalid input"]})
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")
        http_error.__cause__ = validation_error

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, {"field": ["Invalid input"]})

    def test_no_validation_error_returns_reason(self):
        """Test that the reason is returned when no ValidationError is found."""
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, "Unprocessable Entity")

    def test_deeply_nested_validation_error(self):
        """Test extraction when ValidationError is nested deeply."""
        validation_error = ValidationError({"field": ["Invalid input"]})
        level_3 = Exception("Level 3")
        level_3.__cause__ = validation_error
        level_2 = Exception("Level 2")
        level_2.__cause__ = level_3
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")
        http_error.__cause__ = level_2

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, {"field": ["Invalid input"]})

    def test_multiple_exceptions_no_validation_error(self):
        """Test that reason is returned when no ValidationError is in the chain."""
        level_2 = Exception("Level 2")
        level_1 = Exception("Level 1")
        level_1.__cause__ = level_2
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")
        http_error.__cause__ = level_1

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, "Unprocessable Entity")

    def test_validation_error_in_context(self):
        """Test extraction when ValidationError is in __context__ instead of __cause__."""
        validation_error = ValidationError({"field": ["Invalid input"]})
        level_1 = Exception("Level 1")
        level_1.__context__ = validation_error
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")
        http_error.__context__ = level_1

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, {"field": ["Invalid input"]})

    def test_exception_already_visited(self):
        """Test that visited set prevents infinite loops."""
        validation_error = ValidationError({"field": ["Invalid input"]})
        http_error = HTTPUnprocessableEntity(reason="Unprocessable Entity")
        # Create a loop in the exception chain
        validation_error.__cause__ = http_error
        http_error.__cause__ = validation_error

        result = extract_validation_error_message(http_error)
        self.assertEqual(result, {"field": ["Invalid input"]})

    def test_validation_error_as_initial_exception(self):
        """Test when the initial exception is a ValidationError."""
        validation_error = ValidationError({"field": ["Invalid input"]})

        result = extract_validation_error_message(validation_error)
        self.assertEqual(result, {"field": ["Invalid input"]})



"""DID Exchange problem report and reasons."""

import logging
from enum import Enum

from marshmallow import EXCLUDE, ValidationError, validates_schema

from ....problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import PROBLEM_REPORT

HANDLER_CLASS = (
    "acapy_agent.protocols.didexchange.v1_0.handlers."
    "problem_report_handler.DIDXProblemReportHandler"
)

LOGGER = logging.getLogger(__name__)


class ProblemReportReason(Enum):
    """Supported reason codes."""

    INVITATION_NOT_ACCEPTED = "invitation_not_accepted"
    REQUEST_NOT_ACCEPTED = "request_not_accepted"
    REQUEST_PROCESSING_ERROR = "request_processing_error"
    RESPONSE_NOT_ACCEPTED = "response_not_accepted"
    RESPONSE_PROCESSING_ERROR = "response_processing_error"
    COMPLETE_NOT_ACCEPTED = "complete_not_accepted"


class DIDXProblemReport(ProblemReport):
    """Base class representing a connection problem report message."""

    class Meta:
        """DID Exchange problem report metadata."""

        handler_class = HANDLER_CLASS
        message_type = PROBLEM_REPORT
        schema_class = "DIDXProblemReportSchema"


class DIDXProblemReportSchema(ProblemReportSchema):
    """Schema for DIDXProblemReport model class."""

    class Meta:
        """Metadata for connection problem report schema."""

        model_class = DIDXProblemReport
        unknown = EXCLUDE

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields."""

        if not data.get("description", {}).get("code", ""):
            raise ValidationError("Value for description.code must be present")
        elif data.get("description", {}).get("code", "") not in [
            prr.value for prr in ProblemReportReason
        ]:
            locales = list(data.get("description").keys())
            locales.remove("code")
            LOGGER.warning(
                "Unexpected error code received.\n"
                f"Code: {data.get('description').get('code')}, "
                f"Description: {data.get('description').get(locales[0])}"
            )



"""Class to manage jobs in Connection Record."""

from enum import Enum


class TransactionJob(Enum):
    """Represents jobs in Connection Record."""

    TRANSACTION_AUTHOR = (1,)
    TRANSACTION_ENDORSER = (2,)
